# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2012/11/20
# copy: (C) Copyright 2012 Cadit Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import sys, os, yaml
from optparse import OptionParser

from .manager import Manager, FileProvider
from .sender import Sender

#------------------------------------------------------------------------------
def main():

  cmdparser = OptionParser(usage='%prog [options] RECIPIENT ... (--help for details)',
                           )

  cmdparser.add_option('-v', '--verbose',
                       dest='verbose', default=0, action='count',
                       help='enable verbose output to STDERR (multiple invocations'
                       ' increase verbosity)')

  cmdparser.add_option('-t', '--template', metavar='FILENAME',
                       dest='template', default=None, action='store',
                       help='set the email template')

  cmdparser.add_option('-s', '--set-variable', metavar='NAME=VALUE',
                       dest='variables', default=[], action='append',
                       help='set additional variables to the template')

  cmdparser.add_option('-y', '--set-variable-yaml', metavar='NAME=YAML',
                       dest='yamls', default=[], action='append',
                       help='identical to --set-variable, except the value is first passed'
                       ' through a YAML parser')

  cmdparser.add_option('-e', '--add-environment',
                       dest='environ', default=False, action='store_true',
                       help='add all environment variables as template variables')

  cmdparser.add_option('--text',
                       dest='text', default=False, action='store_true',
                       help='send text-only email')

  cmdparser.add_option('--html',
                       dest='html', default=False, action='store_true',
                       help='send html-only email')

  cmdparser.add_option('--raw-html',
                       dest='rawHtml', default=False, action='store_true',
                       help='print out the raw HTML email content')

  cmdparser.add_option('-a', '--all',
                       dest='all', default=False, action='store_true',
                       help='send three emails: text-only, html-only and'
                       ' combined text and html versions')

  cmdparser.add_option('-H', '--smtp-host', metavar='HOST',
                       dest='smtpHost', default=None, action='store',
                       help='set the SMTP host machine to connect to')

  cmdparser.add_option('-P', '--smtp-port', metavar='PORT',
                       dest='smtpPort', default=None, action='store',
                       help='set the SMTP port machine to connect to')

  cmdparser.add_option('-f', '--from', metavar='EADDR',
                       dest='fromAddr', default='noreply@localhost', action='store',
                       help='set the from address (default: %default)')

  cmdparser.add_option('-n', '--no-send',
                       dest='nosend', default=False, action='store_true',
                       help='if specified, the email will not actually be sent'
                       ' but will instead be printed to STDOUT')

  #----------------------------------------------------------------------------
  # process the arguments

  (options, args) = cmdparser.parse_args()

  if options.template is None:
    # TODO: perhaps use STDIN?...
    cmdparser.error('a template is required')

  sender = None
  if options.nosend:
    class NoSender(Sender):
      def send(self, mailfrom, recipients, data):
        print >>sys.stdout, '[  ] mail from:', mailfrom
        print >>sys.stdout, '[  ] recipients:', ', '.join(recipients)
        print >>sys.stdout, '[  ] data:'
        print >>sys.stdout, data
    sender = NoSender()

  prov = FileProvider(os.path.dirname(options.template), os.path.basename(options.template))
  eml  = Manager(sender=sender, provider=prov).newEmail()

  if options.fromAddr is not None:
    eml.setHeader('From', options.fromAddr)

  if options.smtpHost is not None:
    eml.manager.sender.smtpHost = options.smtpHost

  if options.smtpPort is not None:
    eml.manager.sender.smtpPort = options.smtpPort

  if options.environ:
    for k, v in os.environ.items():
      eml[k] = v

  if len(options.variables) > 0:
    for k, v in [v.split('=',1) for v in options.variables]:
      eml[k] = v

  if len(options.yamls) > 0:
    for k, v in [v.split('=',1) for v in options.yamls]:
      eml[k] = yaml.load(v)

  count = 0
  if options.text:    count += 1
  if options.html:    count += 1
  if options.all:     count += 1
  if options.rawHtml: count += 1

  if count > 1:
    cmdparser.error('only one of --text, --html, --all and --raw-html can be used at once')

  if count == 0:
    if options.verbose:
      print >>sys.stderr, '[  ] sending email'
    eml.send(recipients=args)

  if options.all:
    if options.verbose:
      print >>sys.stderr, '[  ] sending combined email'
    eml.send(recipients=args)

  if options.all or options.html:
    if options.verbose:
      print >>sys.stderr, '[  ] sending html-only email'
    eml.include = [i for i in eml.include if i != 'text']
    eml.send(recipients=args)

  if options.all or options.text:
    if options.verbose:
      print >>sys.stderr, '[  ] sending text-only email'
    eml.include = ['text']
    eml.send(recipients=args)

  if options.rawHtml:
    sys.stdout.write(eml.getHtml())

  return 0

#------------------------------------------------------------------------------
if __name__ == '__main__':
  sys.exit(main())

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# lib:  templatealchemy.sqlalchemy
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2013/07/03
# copy: (C) Copyright 2013 Cadit Health Inc., All Rights Reserved.
#------------------------------------------------------------------------------

from __future__ import absolute_import

from StringIO import StringIO
import sqlalchemy as sa
from templatealchemy import api, util

#------------------------------------------------------------------------------
def loadSource(spec=None):
  return SaSource(spec)

#------------------------------------------------------------------------------
class SaSource(api.Source):

  #----------------------------------------------------------------------------
  def __init__(self, spec, name='', tablename=None,
               engine=None, connection=None, table=None, metadata=None,
               *args, **kw):
    super(SaSource, self).__init__(self.ns('sqlalchemy', spec), *args, **kw)
    self.spec = spec
    self.name = name
    self.engine   = engine or sa.create_engine(spec)
    self.conn     = connection or self.engine.connect()
    self.metadata = metadata or sa.MetaData()
    self.table    = table
    if self.table is None:
      self.table = sa.Table(
        tablename or 'template', self.metadata,
        sa.Column('name', sa.String()),
        sa.Column('format', sa.String()),
        sa.Column('content', sa.LargeBinary()),
        )

  #----------------------------------------------------------------------------
  def getSource(self, name):
    return SaSource(
      self.spec, name=self.name + '/' + name if self.name else name,
      engine=self.engine, metadata=self.metadata, table=self.table,
      connection=self.conn,
      )

  #----------------------------------------------------------------------------
  def _get(self, name, format):
    result = self.conn.execute(
      sa.sql
      .select([self.table.c.content])
      .where(sa.sql.and_(self.table.c.name == name,
                         self.table.c.format == format)))
    return StringIO(result.fetchone()[0])

  #----------------------------------------------------------------------------
  def get(self, format):
    return self._get(self.name, format)

  #----------------------------------------------------------------------------
  def getFormats(self):
    result = self.conn.execute(
      sa.sql
      .select([self.table.c.format])
      .where(sa.sql.and_(self.table.c.name == self.name)))
    return [record[0] for record in result.fetchall()]

  #----------------------------------------------------------------------------
  def getRelated(self, name):
    if name.startswith('/'):
      return self._get(name[1:], None)
    return self._get(os.path.dirname(self.name) + '/' + name, None)

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# lib:  templatealchemy.engine
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2013/07/03
# copy: (C) Copyright 2013 Cadit Health Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import os, yaml, re
from . import util
from .util import adict

__all__ = ['Template', 'loadSource', 'loadRenderer']

#------------------------------------------------------------------------------
def loadSource(spec):
  # TODO: make this do more error detection and reporting...
  if spec is None:
    spec = 'pkg'
  if not isinstance(spec, basestring):
    return spec
  spec = spec.split(':', 1)
  loader = util.resolve('templatealchemy.' + spec.pop(0) + '.loadSource')
  return loader(spec.pop(0) if spec else None)

#------------------------------------------------------------------------------
def loadRenderer(spec):
  # TODO: make this do more error detection and reporting...
  if spec is None:
    spec = 'mako'
  if not isinstance(spec, basestring):
    return spec
  spec = spec.split(':', 1)
  loader = util.resolve('templatealchemy.' + spec.pop(0) + '.loadRenderer')
  return loader(spec.pop(0) if spec else None)

#------------------------------------------------------------------------------
def makeLoader(source, relativeTo=None):
  class SourceLoader(yaml.Loader):
    def _include(self, node):
      path = os.path.join(relativeTo or '', node.value)
      return yaml.load(source.getRelated(path).read(),
                       Loader=makeLoader(source, os.path.dirname(path)))
    def _includeRaw(self, node):
      path = os.path.join(relativeTo or '', node.value)
      return source.getRelated(path).read()
  SourceLoader.add_constructor('!include', SourceLoader._include)
  SourceLoader.add_constructor('!include-raw', SourceLoader._includeRaw)
  return SourceLoader

#------------------------------------------------------------------------------
class Template(object):

  # TODO: implement template caching?...

  #----------------------------------------------------------------------------
  def __init__(self, source=None, renderer=None, extmap=None, fmtcmp=None):
    '''
    Construct a new top-level Template with the given `source` and
    `renderer`.

    :Parameters:

    source : { str, templatealchemy.api.Source }

      The persistent read-access storage of template source
      content. The `source` must either be a subclass of
      :class:`templatealchemy.api.Source` or a source specification
      string in the format ``TYPE[:SPEC]``.

    renderer : { str, templatealchemy.api.Renderer }

      The rendering engine for this template. The `renderer` must
      either be a subclass of :class:`templatealchemy.api.Renderer` or
      a renderer specification string in the format ``TYPE[:SPEC]``.

    extmap : dict, optional

      The `extmap` parameter allows remapping of format to extension,
      where needed. For example, if "html" formats for a given
      template are stored on the filesystem with the extension "mako",
      but "text" formats are stored with the extension "mako.txt",
      then the following extmap could be used::

        extmap = { 'html': 'mako', 'text': 'mako.txt' }

      and on the filesystem, for the template
      'pkg:demo:templates/document' you might have the following file
      structure::

        -- demo/templates/
           |-- document.mako
           `-- document.mako.txt

      Note that the special format ``None`` (the symbol, not the
      string) can be used in this dictionary to refer to the default
      template. Generally speaking, it is acceptable to map *to* None,
      but not *from* None to some other value.

    fmtcmp : callable, optional

      Specify the format sorting comparison function, as passed into
      the `cmp` parameter of the `sorted` function. The primary
      consequence of this sorting is that the first format becomes the
      default format. Application-specific logic is required to
      provide a good default format. By default, the formats are
      simply sorted alphabetically.

      @TODO: add smarter default sorting; something like formatted, then
      structured, then semi-structured, then unstructured, then raw::

        xml > xhtml > html > rst > yaml > json > ... > text > data

    '''
    self.source   = loadSource(source)
    self.renderer = loadRenderer(renderer)
    self.context  = adict(template=self)
    self._meta    = None
    self.extmap   = extmap or dict()
    self.rextmap  = {val: key for key, val in self.extmap.items()}
    # TODO: provide a better default sorter -- see pydocs
    self.fmtcmp   = None

  #----------------------------------------------------------------------------
  def getTemplate(self, name):
    # TODO: i need to move to a `Manager` type of approach so that
    #       i only need to pass one parameter to sub-templates.
    return Template(self.source.getSource(name), self.renderer,
                    extmap=self.extmap, fmtcmp=self.fmtcmp)

  #----------------------------------------------------------------------------
  def render(self, format, params):
    if format == 'spec':
      raise TypeError('format "spec" is reserved for internal TemplateAlchemy use')
    format = self.extmap.get(format, format)
    if not format:
      format = self.meta.formats[0]
    return self.renderer.render(self.context, self.source.get(format), params)

  #----------------------------------------------------------------------------
  @property
  def meta(self):
    if self._meta is not None:
      return self._meta
    formats = sorted(self.source.getFormats(), cmp=self.fmtcmp)
    if 'spec' in formats:
      self._meta = yaml.load(self.source.get('spec'), makeLoader(self.source))
      self._meta = adict.__dict2adict__(self._meta, recursive=True)
    else:
      raw = self.source.get(formats[0]).read()
      if '-*- spec -*-' in raw:
        # TODO: generalize how the spec is defined... perhaps let the
        #       active renderer do so?...
        spec = re.search(r'-\*- spec -\*-(.*?)-\*- /spec -\*-', raw, flags=re.DOTALL)
        if spec:
          self._meta = yaml.load(spec.group(1).strip(), makeLoader(self.source))
          self._meta = adict.__dict2adict__(self._meta, recursive=True)
    if self._meta is None:
      self._meta = adict()
    self._meta.formats = [self.rextmap.get(fmt, fmt)
                          for fmt in formats if fmt != 'spec']
    return self._meta

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# lib:  templatealchemy.test
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2013/07/03
# copy: (C) Copyright 2013 Cadit Health Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import unittest, os.path
import templatealchemy
from templatealchemy.util import adict

#------------------------------------------------------------------------------
class TestTemplateAlchemy(unittest.TestCase):

  maxDiff = None

  #----------------------------------------------------------------------------
  def test_mustache(self):
    root = templatealchemy.Template(
      source='pkg:templatealchemy:test_data/mustache',
      renderer='mustache')
    tpl = root.getTemplate('document')
    out = tpl.render('html', adict(
        title='TemplateAlchemy',
        doc=adict(title='Mustache'),
        sections=[
          adict(title='Overview', text='Good'),
          adict(title='Details', text='Poor'),
          adict(title='Utility', text='Excellent'),
          ]))
    chk = '''\
<html>
 <head>
  <title>TemplateAlchemy</title>
 </head>
 <body>
  <h1>TemplateAlchemy</h1>
  <h2>Mustache</h2>
   <h3>Overview</h3>
   <p>Good</p>
   <h3>Details</h3>
   <p>Poor</p>
   <h3>Utility</h3>
   <p>Excellent</p>
 </body>
</html>
'''
    self.assertMultiLineEqual(out, chk)
    self.assertEqual(tpl.meta.formats, ['html'])

  #----------------------------------------------------------------------------
  def test_mako(self):
    root = templatealchemy.Template(
      source='pkg:templatealchemy:test_data/mako',
      renderer='mako',
      extmap={'text': 'txt'})
    tpl = root.getTemplate('document')
    params = adict(
      title='TemplateAlchemy',
      doc=adict(title='Mako'),
      sections=[
        adict(title='Overview', text='Good'),
        adict(title='Details', text='Poor'),
        adict(title='Utility', text='Excellent'),
        ])
    out = tpl.render('html', params)
    chk = '''\
<html>
 <head>
  <title>TemplateAlchemy</title>
 </head>
 <body>
  <h1>TemplateAlchemy</h1>
  <h2>Mako</h2>
   <h3>Overview</h3>
   <p>Good</p>
   <h3>Details</h3>
   <p>Poor</p>
   <h3>Utility</h3>
   <p>Excellent</p>
 </body>
</html>
'''
    self.assertMultiLineEqual(out, chk)
    out = tpl.render('text', params)
    chk = '''\
# TemplateAlchemy

## Mako

### Overview

Good

### Details

Poor

### Utility

Excellent
'''
    self.assertMultiLineEqual(out, chk)
    self.assertEqual(tpl.meta.formats, ['html', 'text'])

  #----------------------------------------------------------------------------
  def test_mako_file(self):
    # TODO: this test will only work if TemplateAlchemy is not zipped...
    #       not quite sure how to test the 'file' source.
    root = templatealchemy.Template(
      source='file:' + os.path.join(os.path.dirname(__file__), 'test_data', 'mako'),
      renderer='mako',
      extmap={'text': 'txt'})
    tpl = root.getTemplate('document')
    params = adict(
      title='TemplateAlchemy',
      doc=adict(title='Mako'),
      sections=[
        adict(title='Overview', text='Good'),
        adict(title='Details', text='Poor'),
        adict(title='Utility', text='Excellent'),
        ])
    out = tpl.render('html', params)
    chk = '''\
<html>
 <head>
  <title>TemplateAlchemy</title>
 </head>
 <body>
  <h1>TemplateAlchemy</h1>
  <h2>Mako</h2>
   <h3>Overview</h3>
   <p>Good</p>
   <h3>Details</h3>
   <p>Poor</p>
   <h3>Utility</h3>
   <p>Excellent</p>
 </body>
</html>
'''
    self.assertMultiLineEqual(out, chk)
    out = tpl.render('text', params)
    chk = '''\
# TemplateAlchemy

## Mako

### Overview

Good

### Details

Poor

### Utility

Excellent
'''
    self.assertMultiLineEqual(out, chk)
    self.assertEqual(tpl.meta.formats, ['html', 'text'])

  #----------------------------------------------------------------------------
  def test_sqlalchemy(self):
    # todo: what if templatealchemy is a zipped archive?...
    path = os.path.join(os.path.dirname(__file__), 'test_data/sa/source.db')
    root = templatealchemy.Template(
      source='sqlalchemy:sqlite:///' + path,
      renderer='mustache')
    tpl = root.getTemplate('document')
    out = tpl.render('html', adict(title='TemplateAlchemy'))
    chk = '<html><body><h1>TemplateAlchemy</h1></body></html>'
    out = tpl.render('text', adict(title='TemplateAlchemy'))
    chk = 'Title: TemplateAlchemy\n'
    self.assertMultiLineEqual(out, chk)
    self.assertEqual(tpl.meta.formats, ['html', 'text'])

  #----------------------------------------------------------------------------
  def test_string(self):
    root = templatealchemy.Template(source='string:ALL YOUR ${plan["from"]} ARE BELONG TO ${plan["to"]}')
    params = dict(plan=adict((('from', 'BASE'), ('to', 'US'))))
    outt = root.render('text', params)
    outh = root.render('html', params)
    self.assertEqual(outt, 'ALL YOUR BASE ARE BELONG TO US')
    self.assertEqual(outh, 'ALL YOUR BASE ARE BELONG TO US')
    self.assertEqual(root.meta.formats, ['data'])

  #----------------------------------------------------------------------------
  def test_commandLine(self):
    from . import cli
    from StringIO import StringIO
    out = StringIO()
    tpl = '''\
<html>
 <head>
  <title>{{title}}</title>
 </head>
 <body>
  <h1>{{title}}</h1>
  <h2>{{doc.title}}</h2>
  {{#sections}}
   <h3>{{title}}</h3>
   <p>{{text}}</p>
  {{/sections}}
 </body>
</html>
'''
    cli.main([
        '--params',
        '{title: "TemplateAlchemy", doc: {title: "Command Line"}, sections: ['
        + '{title: Overview, text: Good},'
        + '{title: Details, text: Poor},'
        + '{title: Utility, text: Excellent},'
        + ']}',
        '--renderer', 'mustache',
        'string:' + tpl], output=out)
    chk = '''\
<html>
 <head>
  <title>TemplateAlchemy</title>
 </head>
 <body>
  <h1>TemplateAlchemy</h1>
  <h2>Command Line</h2>
   <h3>Overview</h3>
   <p>Good</p>
   <h3>Details</h3>
   <p>Poor</p>
   <h3>Utility</h3>
   <p>Excellent</p>
 </body>
</html>
'''
    self.assertMultiLineEqual(out.getvalue(), chk)

  #----------------------------------------------------------------------------
  def test_meta(self):
    root = templatealchemy.Template(
      source='pkg:templatealchemy:test_data/mustache',
      renderer='mustache')
    tpl = root.getTemplate('meta-simple')
    self.assertEqual(tpl.meta.formats, [])
    self.assertTrue('attachments' in tpl.meta)
    self.assertEqual(len(tpl.meta.attachments), 2)
    self.assertEqual(tpl.meta.attachments, [
      dict(name='logo.txt', cid=True, content='My Logo'),
      dict(name='logo.png', cid=True, content='\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\n\x00\x00\x00\n\x01\x03\x00\x00\x00\xb7\xfc]\xfe\x00\x00\x00\x06PLTE\xff\xff\xff\x00\x00\x00U\xc2\xd3~\x00\x00\x00\tpHYs\x00\x00\x0fa\x00\x00\x0fa\x01\xa8?\xa7i\x00\x00\x00\x07tIME\x07\xdd\x07\x07\x12\x04\x11\\\xfd\xd3\x10\x00\x00\x00 IDAT\x08\xd7c\xb0g`\xa8o`x{\x80\xe1\x0c\x18\xdd;\xc0\xf0\xff\x00\x88\x0b\x14\xb4g\x00\x00\xb8(\x0cL\xa6v\x1f\xd8\x00\x00\x00\x00IEND\xaeB`\x82'),
      ])

  #----------------------------------------------------------------------------
  def test_mako_context(self):
    root = templatealchemy.Template(
      source='pkg:templatealchemy:test_data/mako',
      renderer='mako',
      extmap={'text': 'txt'})
    tpl = root.getTemplate('sub')
    out = tpl.render('html', dict(text='some text'))
    chk = '''\
<!DOCTYPE html>
<html>
 <head>
  <title>SubTemplate</title>
 </head>
 <body>
  <h1>Base Common Section</h1>
  

<p>
 Some sub-template content: some text.
</p>

 </body>
</html>
'''
    self.assertMultiLineEqual(out, chk)


#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2012/11/20
# copy: (C) Copyright 2012 Cadit Inc., All Rights Reserved.
#------------------------------------------------------------------------------

'''
An email template abstraction class -- see :class:`caditapp.email.manager`
for details.
'''

#------------------------------------------------------------------------------
# IMPORTS
#------------------------------------------------------------------------------

import re, mimetypes, cssutils, base64, codecs, mako, cssselect
from StringIO import StringIO
import xml.etree.ElementTree as ET
import xpath
from xml.dom import minidom
import email.Encoders, email.Message, email.MIMEMultipart, email.MIMEText
import email.MIMEImage, email.Utils
import html2text

from .idict import idict
from .sender import SmtpSender

#------------------------------------------------------------------------------
# GLOBALS
#------------------------------------------------------------------------------

xmlns = 'http://xmlns.cadit.com/caditapp/email/1.0'
htmlns = 'http://www.w3.org/1999/xhtml'
class MissingHeader(Exception): pass

#------------------------------------------------------------------------------
# TODO: put extractEmails(), getHtmlStyleView() and inlineHtml() in a better place
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# TODO: move this...
#------------------------------------------------------------------------------
# email regex from:
#   http://www.regular-expressions.info/email.html
# and extended to support "...@localhost"
def extractEmails(s):
  if not s:
    return None
  emailRegex = '\\b[A-Z0-9._%+-]+@(?:(?:[A-Z0-9-]+\\.)+[A-Z]{2,4}|localhost)\\b'
  return re.findall(emailRegex, s, re.I)

#------------------------------------------------------------------------------
# TODO: move this...
def getHtmlStyleView(document, css, media='all', name=None,
                     styleCallback=lambda element: None):
  """
  :param document:
    a DOM element (must be minidom-compatible)
  :param css:
    a CSS StyleSheet string
  :param media:
    [optional] TODO: view for which media it should be
  :param name:
    [optional] TODO: names of sheets only
  :param styleCallback:
    [optional] should return css.CSSStyleDeclaration of inline styles,
    for html a style declaration for ``element@style``. Gets one
    parameter ``element`` which is the relevant DOMElement

  returns style view
    a dict of {DOMElement: css.CSSStyleDeclaration} for html

  shamelessly scrubbed from:
    http://cssutils.googlecode.com/svn/trunk/examples/style.py (``getView()``)
    ==> with an adjustment to get around a deprecationwarning in
        cssutils/css/cssstyledeclaration.py:598
    ==> and an adjustment to use "xhtml" translator for the cssselector
  """

  sheet = cssutils.parseString(css)
  css2xpath = cssselect.GenericTranslator()
  view = {}
  specificities = {} # needed temporarily
  # TODO: filter rules simpler?, add @media
  rules = (rule for rule in sheet if rule.type == rule.STYLE_RULE)
  for rule in rules:
    for selector in rule.selectorList:
      xpe = css2xpath.css_to_xpath(selector.selectorText)
      for element in xpath.find(xpe, document):
        if element not in view:
          # add initial empty style declatation
          view[element] = cssutils.css.CSSStyleDeclaration()
          specificities[element] = {}
          # and add inline @style if present
          inlinestyle = styleCallback(element)
          if inlinestyle:
            for p in inlinestyle:
              # set inline style specificity
              view[element].setProperty(p)
              specificities[element][p.name] = (1,0,0,0)
        for p in rule.style:
          # update style declaration
          if p not in view[element]:
            # setProperty needs a new Property object and
            # MUST NOT reuse the existing Property
            # which would be the same for all elements!
            # see Issue #23
            view[element].setProperty(p.name, p.value, p.priority)
            specificities[element][p.name] = selector.specificity
          else:
            sameprio = (p.priority ==
                  view[element].getPropertyPriority(p.name))
            if not sameprio and bool(p.priority) or (
               sameprio and selector.specificity >=
                    specificities[element][p.name]):
              # later, more specific or higher prio
              # NOTE: added explicit removeProperty to get around these warnings:
              #   cssutils-0.9.8a1-py2.6.egg/cssutils/css/cssstyledeclaration.py:598: DeprecationWarning: Call to deprecated method '_getCSSValue'. Use ``property.propertyValue`` instead.
              #   cssutils-0.9.8a1-py2.6.egg/cssutils/css/cssstyledeclaration.py:598: DeprecationWarning: Call to deprecated method '_setCSSValue'. Use ``property.propertyValue`` instead.
              view[element].removeProperty(p.name)
              view[element].setProperty(p.name, p.value, p.priority)
  return view

#------------------------------------------------------------------------------
# TODO: move this to use ET elements to remove the XML roundtrip...
# TODO: this currently strips out the <!DOCTYPE> line if it appears, eg:
#       <!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html4/strict.dtd">
def inlineHtmlStyling_minidom(doc, css):
  def scb(element):
    if element.hasAttribute('style'):
      cssText = element.getAttribute('style')
      return cssutils.css.CSSStyleDeclaration(cssText=cssText)
    return None
  view = getHtmlStyleView(doc, css, styleCallback=scb)
  for element, style in view.items():
    v = style.getCssText(separator=u'')
    element.setAttribute('style', v)
  return doc
def inlineHtmlStyling(etdoc, css):
  html = serializeHtml(etdoc)
  doc  = minidom.parseString(html)
  ret  = inlineHtmlStyling_minidom(doc, css)
  return parseXml(ret.toxml('UTF-8'))

#------------------------------------------------------------------------------
def parseXml(data):
  src = StringIO(data)
  parser = ET.XMLParser()
  parser.parser.UseForeignDTD(True)
    # todo: allow some mechanism for the caller to provide additional entities
  parser.entity['nbsp']  = unichr(160)
  parser.entity['mdash'] = unichr(8212)
  parser.entity['copy']  = unichr(169)
  parser.entity['reg']   = unichr(174)
  etree = ET.ElementTree()
  tree = etree.parse(src, parser=parser)
  return tree

#------------------------------------------------------------------------------
def serializeHtml(xdoc):
  '''
  Serializes ElementTree `xdoc` to HTML.
  NOTE: `xdoc` is mutated to qualify any non-qualified attributes
  and elements with the HTML namespace.
  '''
  # note: qualifying all attributes and elements so that this error is
  # avoided:
  #   ValueError: cannot use non-qualified names with default_namespace option
  # IMHO, it should not be throwing this...
  # http://bugs.python.org/issue17088
  for node in xdoc.iter():
    if not node.tag.startswith('{'):
      node.tag = '{%s}%s' % (htmlns, node.tag)
    for attr in list(node.keys()):
      if not attr.startswith('{'):
        node.set('{%s}%s' % (htmlns, attr), node.get(attr))
        del node.attrib[attr]
  dst = StringIO()
  etree = ET.ElementTree(xdoc)
  etree.write(dst, default_namespace=htmlns)
  # todo: it would be nice if the HTML output were 'pretty-printed'...
  return dst.getvalue()

#------------------------------------------------------------------------------
# shamelessly scrubbed from:
#   http://www.w3schools.com/tags/ref_symbols.asp
# with additions for &copy; and &reg;
r2a_map = dict((

  (u'\u00a9', '(C)'),                  # © &#169; &copy; copyright symbol
  (u'\u00ae', '(R)'),                  # ® &#174; &reg; registered trademark

  # math symbols supported by HTML
  (u'\u2200', '<for-all>'),            # ∀ &#8704; &forall; for all
  (u'\u2202', '<part>'),               # ∂ &#8706; &part; part
  (u'\u2203', '<exists>'),             # ∃ &#8707; &exist; exists
  (u'\u2205', '<empty>'),              # ∅ &#8709; &empty; empty
  (u'\u2207', '<nabla>'),              # ∇ &#8711; &nabla; nabla
  (u'\u2208', '<isin>'),               # ∈ &#8712; &isin; isin
  (u'\u2209', '<notin>'),              # ∉ &#8713; &notin; notin
  (u'\u220b', '<ni>'),                 # ∋ &#8715; &ni; ni
  (u'\u220f', '<prod>'),               # ∏ &#8719; &prod; prod
  (u'\u2211', '<sum>'),                # ∑ &#8721; &sum; sum
  (u'\u2212', '<minus>'),              # − &#8722; &minus; minus
  (u'\u2217', '<lowast>'),             # ∗ &#8727; &lowast; lowast
  (u'\u221a', '<square-root>'),        # √ &#8730; &radic; square root
  (u'\u221d', '<proportional-to>'),    # ∝ &#8733; &prop; proportional to
  (u'\u221e', '<infinity>'),           # ∞ &#8734; &infin; infinity
  (u'\u2220', '<angle>'),              # ∠ &#8736; &ang; angle
  (u'\u2227', '<and>'),                # ∧ &#8743; &and; and
  (u'\u2228', '<or>'),                 # ∨ &#8744; &or; or
  (u'\u2229', '<cap>'),                # ∩ &#8745; &cap; cap
  (u'\u222a', '<cup>'),                # ∪ &#8746; &cup; cup
  (u'\u222b', '<integral>'),           # ∫ &#8747; &int; integral
  (u'\u2234', '<therefore>'),          # ∴ &#8756; &there4; therefore
  (u'\u223c', '<similar-to>'),         # ∼ &#8764; &sim; similar to
  (u'\u2245', '<congruent-to>'),       # ≅ &#8773; &cong; congruent to
  (u'\u2248', '<almost-equal>'),       # ≈ &#8776; &asymp; almost equal
  (u'\u2260', '<not-equal>'),          # ≠ &#8800; &ne; not equal
  (u'\u2261', '<equivalent>'),         # ≡ &#8801; &equiv; equivalent
  (u'\u2264', '<less-or-equal>'),      # ≤ &#8804; &le; less or equal
  (u'\u2265', '<greater-or-equal>'),   # ≥ &#8805; &ge; greater or equal
  (u'\u2282', '<subset-of>'),          # ⊂ &#8834; &sub; subset of
  (u'\u2283', '<superset-of>'),        # ⊃ &#8835; &sup; superset of
  (u'\u2284', '<not-subset-of>'),      # ⊄ &#8836; &nsub; not subset of
  (u'\u2286', '<subset-or-equal>'),    # ⊆ &#8838; &sube; subset or equal
  (u'\u2287', '<superset-or-equal>'),  # ⊇ &#8839; &supe; superset or equal
  (u'\u2295', '<circled-plus>'),       # ⊕ &#8853; &oplus; circled plus
  (u'\u2297', '<circled-times>'),      # ⊗ &#8855; &otimes; circled times
  (u'\u22a5', '<perpendicular>'),      # ⊥ &#8869; &perp; perpendicular
  (u'\u22c5', '<dot-operator>'),       # ⋅ &#8901; &sdot; dot operator

  # greek letters supported by HTML
  (u'\u0391', '<Alpha>'),              # Α &#913; &Alpha; Alpha
  (u'\u0392', '<Beta>'),               # Β &#914; &Beta; Beta
  (u'\u0393', '<Gamma>'),              # Γ &#915; &Gamma; Gamma
  (u'\u0394', '<Delta>'),              # Δ &#916; &Delta; Delta
  (u'\u0395', '<Epsilon>'),            # Ε &#917; &Epsilon; Epsilon
  (u'\u0396', '<Zeta>'),               # Ζ &#918; &Zeta; Zeta
  (u'\u0397', '<Eta>'),                # Η &#919; &Eta; Eta
  (u'\u0398', '<Theta>'),              # Θ &#920; &Theta; Theta
  (u'\u0399', '<Iota>'),               # Ι &#921; &Iota; Iota
  (u'\u039a', '<Kappa>'),              # Κ &#922; &Kappa; Kappa
  (u'\u039b', '<Lambda>'),             # Λ &#923; &Lambda; Lambda
  (u'\u039c', '<Mu>'),                 # Μ &#924; &Mu; Mu
  (u'\u039d', '<Nu>'),                 # Ν &#925; &Nu; Nu
  (u'\u039e', '<Xi>'),                 # Ξ &#926; &Xi; Xi
  (u'\u039f', '<Omicron>'),            # Ο &#927; &Omicron; Omicron
  (u'\u03a0', '<Pi>'),                 # Π &#928; &Pi; Pi
  (u'\u03a1', '<Rho>'),                # Ρ &#929; &Rho; Rho
  (u'\u03a3', '<Sigma>'),              # Σ &#931; &Sigma; Sigma
  (u'\u03a4', '<Tau>'),                # Τ &#932; &Tau; Tau
  (u'\u03a5', '<Upsilon>'),            # Υ &#933; &Upsilon; Upsilon
  (u'\u03a6', '<Phi>'),                # Φ &#934; &Phi; Phi
  (u'\u03a7', '<Chi>'),                # Χ &#935; &Chi; Chi
  (u'\u03a8', '<Psi>'),                # Ψ &#936; &Psi; Psi
  (u'\u03a9', '<Omega>'),              # Ω &#937; &Omega; Omega
  (u'\u03b1', '<alpha>'),              # α &#945; &alpha; alpha
  (u'\u03b2', '<beta>'),               # β &#946; &beta; beta
  (u'\u03b3', '<gamma>'),              # γ &#947; &gamma; gamma
  (u'\u03b4', '<delta>'),              # δ &#948; &delta; delta
  (u'\u03b5', '<epsilon>'),            # ε &#949; &epsilon; epsilon
  (u'\u03b6', '<zeta>'),               # ζ &#950; &zeta; zeta
  (u'\u03b7', '<eta>'),                # η &#951; &eta; eta
  (u'\u03b8', '<theta>'),              # θ &#952; &theta; theta
  (u'\u03b9', '<iota>'),               # ι &#953; &iota; iota
  (u'\u03ba', '<kappa>'),              # κ &#954; &kappa; kappa
  (u'\u03bb', '<lambda>'),             # λ &#955; &lambda; lambda
  (u'\u03bc', '<mu>'),                 # μ &#956; &mu; mu
  (u'\u03bd', '<nu>'),                 # ν &#957; &nu; nu
  (u'\u03be', '<xi>'),                 # ξ &#958; &xi; xi
  (u'\u03bf', '<omicron>'),            # ο &#959; &omicron; omicron
  (u'\u03c0', '<pi>'),                 # π &#960; &pi; pi
  (u'\u03c1', '<rho>'),                # ρ &#961; &rho; rho
  (u'\u03c2', '<sigmaf>'),             # ς &#962; &sigmaf; sigmaf
  (u'\u03c3', '<sigma>'),              # σ &#963; &sigma; sigma
  (u'\u03c4', '<tau>'),                # τ &#964; &tau; tau
  (u'\u03c5', '<upsilon>'),            # υ &#965; &upsilon; upsilon
  (u'\u03c6', '<phi>'),                # φ &#966; &phi; phi
  (u'\u03c7', '<chi>'),                # χ &#967; &chi; chi
  (u'\u03c8', '<psi>'),                # ψ &#968; &psi; psi
  (u'\u03c9', '<omega>'),              # ω &#969; &omega; omega
  (u'\u03d1', '<theta-symbol>'),       # ϑ &#977; &thetasym; theta symbol
  (u'\u03d2', '<upsilon-symbol>'),     # ϒ &#978; &upsih; upsilon symbol
  (u'\u03d6', '<pi-symbol>'),          # ϖ &#982; &piv; pi symbol

  # other symbols supported by HTML
  (u'\u0152', 'OE'),                   # Œ &#338; &OElig; capital ligature OE
  (u'\u0153', 'oe'),                   # œ &#339; &oelig; small ligature oe
  # TODO: should these be decorated?...
  (u'\u0160', 'S'),                    # Š &#352; &Scaron; capital S with caron
  (u'\u0161', 's'),                    # š &#353; &scaron; small S with caron
  (u'\u0178', 'Y'),                    # Ÿ &#376; &Yuml; capital Y with diaeres
  (u'\u0192', '?'),                    # ƒ &#402; &fnof; f with hook

  (u'\u02c6', '^'),                    # ˆ &#710; &circ; modifier letter circumflex accent
  (u'\u02dc', '~'),                    # ˜ &#732; &tilde; small tilde
  (u'\u2002', ' '),                    #   &#8194; &ensp; en space
  (u'\u2003', ' '),                    #   &#8195; &emsp; em space
  (u'\u2009', ' '),                    #   &#8201; &thinsp; thin space
  (u'\u200c', '?'),                    # ‌ &#8204; &zwnj; zero width non-joiner
  (u'\u200d', '?'),                    # ‍ &#8205; &zwj; zero width joiner
  (u'\u200e', '?'),                    # ‎ &#8206; &lrm; left-to-right mark
  (u'\u200f', '?'),                    # ‏ &#8207; &rlm; right-to-left mark
  (u'\u2013', '-'),                    # – &#8211; &ndash; en dash
  (u'\u2014', '-'),                    # — &#8212; &mdash; em dash
  (u'\u2018', '\''),                   # ‘ &#8216; &lsquo; left single quotation mark
  (u'\u2019', '\''),                   # ’ &#8217; &rsquo; right single quotation mark
  (u'\u201a', '\''),                   # ‚ &#8218; &sbquo; single low-9 quotation mark
  (u'\u201c', '"'),                    # “ &#8220; &ldquo; left double quotation mark
  (u'\u201d', '"'),                    # ” &#8221; &rdquo; right double quotation mark
  (u'\u201e', '"'),                    # „ &#8222; &bdquo; double low-9 quotation mark
  (u'\u2020', '?'),                    # † &#8224; &dagger; dagger
  (u'\u2021', '?'),                    # ‡ &#8225; &Dagger; double dagger
  (u'\u2022', '*'),                    # • &#8226; &bull; bullet
  (u'\u2026', '...'),                  # … &#8230; &hellip; horizontal ellipsis
  (u'\u2030', '0/00'),                 # ‰ &#8240; &permil; per mille
  (u'\u2032', '\''),                   # ′ &#8242; &prime; minutes
  (u'\u2033', '"'),                    # ″ &#8243; &Prime; seconds
  (u'\u2039', '<'),                    # ‹ &#8249; &lsaquo; single left angle quotation
  (u'\u203a', '>'),                    # › &#8250; &rsaquo; single right angle quotation
  (u'\u203e', '-'),                    # ‾ &#8254; &oline; overline
  (u'\u20ac', '<euro>'),               # € &#8364; &euro; euro
  (u'\u2122', '<TM>'),                 # ™ &#8482; &trade; trademark
  (u'\u2190', '<-'),                   # ← &#8592; &larr; left arrow
  (u'\u2191', '?'),                    # ↑ &#8593; &uarr; up arrow
  (u'\u2192', '->'),                   # → &#8594; &rarr; right arrow
  (u'\u2193', '?'),                    # ↓ &#8595; &darr; down arrow

  # additional right arrows scrubbed from:
  #   http://right-arrow.net/
  (u'\u21d2', '=>'),                   # ⇒ &#8658; &rArr;
  (u'\u21ac', '->'),                   # ↬ &#8620;
  (u'\u21c9', '=>'),                   # ⇉ &#8649;
  (u'\u21e8', '=>'),                   # ⇨ &#8680;
  (u'\u21e5', '->|'),                  # ⇥ &#8677;
  (u'\u21e2', '->'),                   # ⇢ &#8674;
  (u'\u21aa', '->'),                   # ↪ &#8618;
  (u'\u21a6', '|->'),                  # ↦ &#8614;
  (u'\u21f6', '=>'),                   # ⇶ &#8694;
  (u'\u21a3', '>->'),                  # ↣ &#8611;
  (u'\u21f0', '|=>'),                  # ⇰ &#8688;
  (u'\u219d', '~>'),                   # ↝ &#8605;
  (u'\u21a0', '->>'),                  # ↠ &#8608;
  (u'\u21fe', '->'),                   # ⇾ &#8702;

  # others...
  (u'\u21e6', '<='),                   # ⇦ &#8678;

  ))

def reduce2ascii(text):
  ret = []
  for c in text:
    if ord(c) < 128:
      ret.append(c)
      continue
    # todo: handle ascii chars in the 128 -> 255 range...
    ret.append(r2a_map.get(c, '?'))
  return ''.join(ret)
  #return text.replace(u'\u21d2', '=>')

#------------------------------------------------------------------------------
_removeCidsRe = re.compile('\\!\\[([^\\]]*)\\]\\(cid:[^)]*\\)')
def removeCids(text):
  return _removeCidsRe.sub('[\\1]', text)

#------------------------------------------------------------------------------
# todo: move this?
smtp_header_element_upper = ['cc', 'id', 'spf']
def smtpHeaderFormat(h):
  return '-'.join([e.upper() if e in smtp_header_element_upper else e.title()
                   for e in [e.lower() for e in h.split('-')]])

#------------------------------------------------------------------------------
# todo: installing a global unicode decode/encode translation error handler...
#       this is not ideal... see Email.getSubject() for details...
#------------------------------------------------------------------------------
def caditapp_email_unicode2ascii(err):
  if err.reason != 'ordinal not in range(128)':
    raise err
  # todo: shouldn't this just use `r2a_map`?...
  if err.object[err.start] == unichr(160):
    # &nbsp;
    return (u' ', err.start + 1)
  if err.object[err.start] == unichr(8212):
    # &mdash;
    return (u'--', err.start + 1)
  return (u'?', err.start + 1)
#------------------------------------------------------------------------------
try:
  codecs.lookup_error('caditapp_email_unicode2ascii')
except LookupError:
  codecs.register_error('caditapp_email_unicode2ascii', caditapp_email_unicode2ascii)

extractEnv_expr = re.compile(
  '<email:env\s+name\s*=\s*"([^"]*)"\s*>(.*?)</email:env>', flags=re.DOTALL)

#------------------------------------------------------------------------------
class Email(object):
  '''
  The ``caditapp.email.Email`` object provides access to an email
  template, such that formally specified emails can be easily
  generated.

  Email headers (i.e. ``To``, ``From``, ``Received``, etc) can be set
  via the :meth:`setHeader` method. It is recommended to set at least
  the ``To`` and ``From`` headers, either via the API or from the
  template (as documented in the :mod:`caditapp.email` module).

  Variables used by the underlying mako template rendering are
  specified using Email object item attributes, for example::

    email = caditapp.email.Manager(...).newEmail(...)
    email['givenname'] = 'John'
    email['surname'] = 'Doe'
    email.setHeader('To', 'john.doe@example.com')
    email.setHeader('From', 'noreply@example.com')
    email.send()
  '''

  #----------------------------------------------------------------------------
  def __init__(self, manager, provider=None, name=None):
    self.manager            = manager
    self.provider           = provider or manager.provider
    self.name               = name
    self.style              = None
    self.include            = ['text', 'html', 'attachments']
    self.textRenderEncoding = self.manager.defaultTextRenderEncoding
    self.htmlRenderEncoding = self.manager.defaultHtmlRenderEncoding
    self.textEncoding       = self.manager.defaultTextEncoding
    self.htmlEncoding       = self.manager.defaultHtmlEncoding
    self.transferEncoding   = self.manager.defaultTransferEncoding
    self.encoding           = self.manager.defaultEncoding
    self.maxSubjectLength   = self.manager.defaultMaxSubjectLength
    self.snipIndicator      = self.manager.defaultSnipIndicator
    self.headers            = idict()
    self.params             = {}
    self.attachments        = []
    self.boundary           = self.manager.defaultBoundary
    self.template           = self.provider.getTemplate(self.name)
    self.cidRewrite         = None
    self.attachmentTable    = None
    for key, val in self.provider.getMap('headers', {}).items():
      self.headers[key] = val

  #----------------------------------------------------------------------------
  def __setitem__(self, key, val): self.params[key] = val
  def __getitem__(self, key): return self.params[key]
  def __delitem__(self, key): del self.params[key]

  #----------------------------------------------------------------------------
  def setHeader(self, key, val):
    'Set the email header `key` to `val`, overriding all other header sources.'
    self.headers[key] = val

  #----------------------------------------------------------------------------
  def getHeader(self, key):
    '''
    Get the current value of the header `key`. Note that this will only return
    the value as set with a prior :meth:`setHeader` call - it will *not*
    extract headers set by the template(s).
    '''
    return self.headers[key]

  #----------------------------------------------------------------------------
  def hasHeader(self, key):
    '''
    Returns whether or not the header `key` has been set with a prior call
    to :meth:`setHeader` - it will *not* extract headers set by the template(s).
    '''
    return self.headers.has_key(key)

  #----------------------------------------------------------------------------
  def delHeader(self, key):
    '''
    If the header `key` had been set with a prior call to
    :meth:`setHeader`, this will unset the header. Note that it will
    *not* delete any headers set by the template(s).
    '''
    del self.headers[key]

  #----------------------------------------------------------------------------
  def getEnvs(self):
    '''
    Returns a dictionary of parameters set by the template (this is so
    that templates can communicate non-output parameters back to the
    calling application).
    '''
    # TODO: this is all wrong. need to re-architect this.
    ret = dict()
    for match in extractEnv_expr.finditer(self.template.source):
      # todo: ugh. i need to do XML un-escaping too...
      ret[match.group(1)] = match.group(2)
    return ret

  #----------------------------------------------------------------------------
  def getEnv(self, key, default=None):
    '''
    Returns the template-set parameter name `key`. If not set, `default`
    is returned. See :meth:`getEnvs` for details.
    '''
    return self.getEnvs().get(key, default)

  #----------------------------------------------------------------------------
  def addAttachment(self, name, value, cid=False, contentType=None):
    '''
    Add an attachment to the email. The `name` is the default filename that
    will be offered when the recipient tries to "Save..." the attachment. The
    `value` is the actual content of the attachment. If `cid` is True, then
    the attachment will be stored as an embedded object, and will therefore
    not be directly saveable by the recipient and it can be accessed from
    within the HTML, for example, with the `name` set to ``'logo.png'``, the
    following will result in a valid image reference in the HTML::

      <img alt="The Logo" src="cid:logo.png"/>

    (note the ``cid:`` prefix.)
    '''
    att = dict(name=name, value=value, cid=cid)
    if contentType is not None:
      att['content-type'] = contentType
    self.attachments.append(att)

  #----------------------------------------------------------------------------
  def setStyle(self, css):
    '''
    Add the specified CSS style instructions to the current HTML rendering.
    Note the following current limitations of the underlying CSS engine:

      * element names must be lowercase, e.g. "p {...}", not "P {....}"
      * "body" element cannot be targeted, e.g. "body {...}" does NOT work
    '''
    self.style = css

  #----------------------------------------------------------------------------
  def getTemplateXml(self):
    # todo: use TemplateLookup?... if so, make sure to use the manager's...
    #         ie something like ``self.manager.makoLookup``...
    # todo: filesystem_checks=False configurable
    return parseXml(self.template.render(**self.params))

  #----------------------------------------------------------------------------
  def getTemplateHeaders(self):
    ret = {}
    # todo: this requires that headers be defined as separate elements, eg:
    #         <email:header name="..." value="...">...</email:header>
    #       rather than being able to tag an existing node, such as:
    #         <p>This email was sent to <span email:header="To">...</span>.</p>
    xdoc = self.getTemplateXml()
    etag = '{%s}header' % (xmlns,)
    for node in xdoc.iter():
      if node.tag == etag:
        ret[node.get('name')] = node.get('value') or node.text
        continue
      if node.get(etag) is not None:
        ret[node.get(etag)] = node.text
    return ret

  #----------------------------------------------------------------------------
  def getTemplateAttachments(self):
    ret = []
    xdoc = self.getTemplateXml()
    for node in xdoc.iter('{%s}attachment' % (xmlns,)):
      attm  = dict(name = node.get('name'),
                   cid  = node.get('cid', 'false').lower() == 'true')
      value = node.get('value', None) or node.text
      if node.get('encoding', None) == 'base64':
        value = base64.b64decode(value)
      attm['value'] = value
      ct = node.get('content-type', None)
      if ct is not None:
        attm['content-type'] = ct
      ret.append(attm)
    return ret

  #----------------------------------------------------------------------------
  def getTemplateStyle(self):
    ret  = []
    xdoc = self.getTemplateXml()
    # todo: what if the node has a "src" attribute...
    for node in xdoc.findall('{%s}head/{%s}style[@type="text/css"]' % (htmlns, htmlns)):
      ret.append(node.text)
    return ' '.join(ret)

  #----------------------------------------------------------------------------
  def getHtml(self, standalone=False):
    '''
    Returns the raw HTML that would be generated if the email were
    sent with the current settings. Use :meth:`send` to actually send
    the email.
    '''

    if standalone:
      ret = self.getHtml()
      return self.inlineCidAttachments(ret)

    tpl = self.provider.getTemplate('html', self.template)
    html = tpl.render(**self.params)

    # todo: this double roundtrip of parse/serialize html is ridiculous

    # remove all caditapp/email xmlns elements and attributes and non-inline css
    # note: use list() so that i can mutate the underlying object
    html = parseXml(html)
    def stripSpecial(el):
      for attr in list(el.keys()):
        if attr.startswith('{%s}' % (xmlns,)):
          del el.attrib[attr]
      for node in list(el.getchildren()):
        if node.tag.startswith('{%s}' % (xmlns,)):
          el.remove(node)
          continue
        stripSpecial(node)
    stripSpecial(html)
    for topnode in html.getchildren():
      if topnode.tag == '{%s}head' % (htmlns,):
        for node in list(topnode.getchildren()):
          if node.tag == '{%s}style' % (htmlns,) and node.get('type') == 'text/css':
            topnode.remove(node)

    style = (self.getTemplateStyle() or '').strip()
    if self.style is not None and len(self.style) > 0:
      style += self.style
    if len(style.strip()) > 0:
      html = inlineHtmlStyling(html, style)

    return serializeHtml(html)

  #----------------------------------------------------------------------------
  def inlineCidAttachments(self, html):
    # TODO: this is a "brute-force" approach... clean it up!
    atts = self.getAttachments()
    for att in atts:
      if not att.get('cid', False):
        continue
      # TODO: assuming PNG... inspect att['content-type']...
      value = 'data:image/png;base64,' + base64.b64encode(att['value'])
      html = html.replace('cid:' + att['name'], value)
    return html

  #----------------------------------------------------------------------------
  def getText(self):
    '''
    Returns the plain-text version of the email that would be
    generated if it were sent with the current settings. Use
    :meth:`send` to actually send the email.
    '''
    tpl = self.provider.getTemplate('text')
    if tpl is not None:
      return removeCids(reduce2ascii(tpl.render(**self.params)))
    html = self.getHtml()
    # todo: it would be interesting to be able to configure html2text to only
    #       put it footnotes for IMG tags that had non-"cid:" image references...
    text = html2text.html2text(html)
    # TODO: html2text should have taken care of this... but since it hasn't,
    #       are there any other characters that need to be escaped?...
    return removeCids(reduce2ascii(text))

  subject_collapse_spaces = re.compile(r'[\s]+', re.DOTALL)
  subject_remove_nonascii = re.compile(r'[^ -~]+', re.DOTALL)

  #----------------------------------------------------------------------------
  def getSubject(self):
    '''
    Returns the ``Subject`` header of the email that would be
    generated if it were sent with the current settings. Use
    :meth:`send` to actually send the email.
    '''
    tpl = self.provider.getTemplate('subject', self.template)
    # note: purposefully NOT using encoding=self.encoding so that i can
    #       then more cleanly replace HTML entity characters...
    # todo: clean up this entity-replacement strategy
    # todo: if no subject is found, the getText() version should be reduced
    #      and truncated...

    xdoc = parseXml(tpl.render(**self.params))
    etag = '{%s}subject' % (xmlns,)
    ret = []
    for node in xdoc.iter():
      if node.tag == etag or node.get(etag) == 'content':
        ret.append(node.text)
    ret = ' '.join(ret).encode('us-ascii', 'caditapp_email_unicode2ascii')
    ret = self.subject_collapse_spaces.sub(' ', ret)
    ret = self.subject_remove_nonascii.sub('', ret)
    if self.maxSubjectLength is not None and self.snipIndicator is not None:
      if len(ret) > self.maxSubjectLength:
        ret = ret[:self.maxSubjectLength - len(self.snipIndicator)] + self.snipIndicator
    return reduce2ascii(ret)

  #----------------------------------------------------------------------------
  def getProviderAttachments(self):
    # TODO: clean this up...
    if self.attachmentTable is None:
      self.attachmentTable = self.provider.getMap('attachments', {})
    return self.attachmentTable

  #----------------------------------------------------------------------------
  def getAttachments(self):
    atts = self.attachments[:]
    for att in self.getTemplateAttachments():
      if att['name'] not in [a['name'] for a in atts]:
        atts.append(att)
    for att in self.getProviderAttachments():
      if att['name'] not in [a['name'] for a in atts]:
        atts.append(att)
    return atts

  #----------------------------------------------------------------------------
  def getMimeAttachments(self):

    atts = self.getAttachments()

    # TODO: these attachments have been geared toward inlined image
    #       attachments... is this how standard attachments are sent as well?

    ret = []
    for adef in atts:
      # todo: a lot more escaping might need to be done here... especially of
      #       the attachment attributes...
      name  = adef['name']
      value = adef['value']
      ctype = adef.get('content-type', None) \
              or mimetypes.guess_type(name or '', False)[0] \
              or 'application/octet-stream'
      maintype, subtype = ctype.split('/', 1)
      if maintype == 'text':
        # Note: we should handle calculating the charset
        att = email.MIMEText.MIMEText(value, _subtype=subtype)
      elif maintype == 'image':
        att = email.MIMEImage.MIMEImage(value, _subtype=subtype, name=name)
      elif maintype == 'audio':
        att = email.MIMEAudio.MIMEAudio(value, _subtype=subtype)
      else:
        att = email.MIMEBase.MIMEBase(maintype, subtype)
        att.set_payload(value)
        email.Encoders.encode_base64(att)
      if adef.get('cid', False):
        att.add_header('Content-Disposition', 'attachment')
        att.add_header('Content-ID', '<' + name + '>')
      else:
        att.add_header('Content-Disposition', 'attachment', filename=name)
      ret.append(att)
    return ret

  #----------------------------------------------------------------------------
  def getOutputHeaders(self):
    curheaders = idict()
    curheaders.update(self.getTemplateHeaders())
    curheaders.update(self.headers)
    defaultHeaders = {
      'Subject':     lambda: self.getSubject(),
      }
    for name, call in defaultHeaders.items():
      if curheaders.has_key(name):
        continue
      curheaders[name] = call()
    if self.manager.updateHeaders is not None:
      self.manager.updateHeaders(self, curheaders)
    return curheaders

  #----------------------------------------------------------------------------
  def getSmtpData(self):
    '''
    Returns the raw SMTP data that would be sent if the email were sent
    with the current settings. Use :meth:`send` to actually send the email.
    '''
    return self._getSmtpData(self.getOutputHeaders())

  #----------------------------------------------------------------------------
  def _getSmtpData(self, curheaders):

    # TODO: is this the right structure for adding non-Content-ID attachments?...
    #       (e.g. boarding-pass style appt sheet)

    #  mail structure:
    #
    #     multipart/alternative
    #     |-- text/plain
    #     |-- multipart/related; type="text/html"
    #     |   |-- text/html
    #     |   `-- image/png... [attachments with "Content-ID"]
    #     `-- application/octet-stream... [attachments without "Content-ID"]
    #
    #               Content-Type: image/png
    #               Content-Transfer-Encoding: base64
    #               Content-Location: file:///.../...
    #               Content-ID: <ImageName>

    curheaders = idict(curheaders)

    # reduce headers to those that should be in outbound email
    # todo: should i instead be filtering for known allowed headers instead
    #       of removing known disallowed headers?...
    for header in ('bcc',):
      if curheaders.get(header):
        del curheaders[header]

    # the following text/html encoding selection is done only because
    # utf-8 results in base64 transfer encoding, which sucks because it
    # is not human-readable... hence trying others first. and the problem
    # with iso-8859-1 is that the "&copy;" and "&nbsp;" entities don't
    # show up, but it uses quoted-printable...
    # ==> UPDATE: i can't seem to reproduce this... the following translations
    #             occur: &mdash; => &#8212;, &copy; => &#169;, &nbsp; => &#160;

    # TODO: address the other self.encoding uses and see if the default
    #       should really be us-ascii instead of None...

    # encode the text-only alternative
    txtenc  = self.textEncoding or self.transferEncoding or self.encoding
    if txtenc is not None:
      txt = email.MIMEText.MIMEText(self.getText(), 'plain', txtenc)
    else:
      # do several encoding round trips until one is found that does not
      # degrade the content
      for txtenc in [ 'ascii', 'iso-8859-1', 'utf-8' ]:
        src = self.getText()
        try: txt = email.MIMEText.MIMEText(src, 'plain', txtenc)
        except UnicodeEncodeError: continue
        if src == txt.get_payload(decode=True):
          break

    # encode the html alternative
    htmlenc  = self.htmlEncoding or self.transferEncoding or self.encoding
    if htmlenc is not None:
      html = email.MIMEText.MIMEText(self.getHtml(), 'html', htmlenc)
    else:
      # do several encoding round trips until one is found that does not
      # degrade the content
      for htmlenc in [ 'ascii', 'iso-8859-1', 'utf-8' ]:
        src = self.getHtml()
        try: html = email.MIMEText.MIMEText(src, 'html', htmlenc)
        except UnicodeEncodeError: continue
        if src == html.get_payload(decode=True):
          break

    # and now stitch all the parts together

    mbound  = self.boundary and ( '==' + self.boundary + '==' ) or None
    msg = email.MIMEMultipart.MIMEMultipart('alternative', boundary=mbound)

    rbound  = self.boundary and ( '==' + self.boundary + '-related==' ) or None
    related = email.MIMEMultipart.MIMEMultipart('related', boundary=rbound)

    if 'text' in self.include:
      msg.attach(txt)

    if 'html' in self.include:
      related.attach(html)
      msg.attach(related)

    if 'attachments' in self.include:
      for att in self.getMimeAttachments():
        if 'Content-ID' in att.keys():
          related.attach(att)
        else:
          msg.attach(att)

    if self.include == ['text']:
      for k,v in curheaders.items():
        txt[smtpHeaderFormat(k)] = v
      return txt.as_string()

    for k,v in curheaders.items():
      msg[smtpHeaderFormat(k)] = v

    return msg.as_string()

  #----------------------------------------------------------------------------
  def send(self, mailfrom=None, recipients=None):
    '''
    Send the email. The ``To`` header can be overriden with the `recipients`
    parameter, and the ``From`` header con be overriden with the `mailfrom`
    parameter. Note that if these parameters are used, then the headers are
    NOT adjusted to reflect the new settings (this is to allow ``Bcc`` type
    behavior).

    Parameters:

    :param mailfrom:

      string, email address to set as the ``MAIL FROM`` address on the
      SMTP protocol level. Note that this will NOT set or override the
      active ``From`` header. If not specified, the address will be
      extracted from the ``From`` header.

    :param recipients:

      string or list-of-string, email addresses to send this email to,
      on the SMTP protocol level. Note that this will NOT set or
      override the active ``To`` header. If not specified, the
      addresses will be extracted from the ``To``, ``CC``, and ``BCC``
      headers.
    '''
    hdrs = self.getOutputHeaders()
    if mailfrom is None:
      mailfrom = extractEmails(hdrs.get('from'))
      if not mailfrom:
        raise MissingHeader('email source ("from") not specified')
      mailfrom = mailfrom[0]
    if recipients is None:
      recipients = extractEmails(hdrs.get('to')) or []
      recipients += extractEmails(hdrs.get('cc')) or []
      recipients += extractEmails(hdrs.get('bcc')) or []
      if not recipients or len(recipients) <= 0:
        raise MissingHeader('email destination ("to") not specified')
    elif isinstance(recipients, basestring):
      recipients = [recipients]
    if hasattr(self.manager, 'updateRecipients'):
      recipients = self.manager.updateRecipients(self, recipients)
    if hasattr(self.manager, 'updateMailfrom'):
      mailfrom = self.manager.updateMailfrom(self, mailfrom)
    self.manager.sender.send(mailfrom, recipients, self._getSmtpData(hdrs))

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2012/11/20
# copy: (C) Copyright 2012 Cadit Inc., All Rights Reserved.
#------------------------------------------------------------------------------

# todo: convert all MultiLineEqual()'s that compare XML to normalize
#       the XML first.

import sys, unittest, base64, difflib, re

from .manager import Manager, StringProvider
from .sender import Sender, StoredSender
from . import emailprov

def eqstrip(val):
  return val.replace('\n', '').replace('\r\n', '').replace(' ', '')

#------------------------------------------------------------------------------
class TestEmail(unittest.TestCase):

  maxDiff = None

  #----------------------------------------------------------------------------
  def test_inlineHtmlStyling(self):
    html = '<html><body><div>foo</div></body></html>'
    css  = 'body{color:red}body > div{font-size:10px}'
    chk  = '''<html xmlns="http://www.w3.org/1999/xhtml">
    <body style="color: red">
      <div style="font-size: 10px">foo</div>
    </body>
  </html>
'''
    xdoc = emailprov.parseXml(html)
    xdoc = emailprov.inlineHtmlStyling(xdoc, css)
    out  = emailprov.serializeHtml(xdoc)
    self.assertEqual(eqstrip(out), eqstrip(chk))

  #----------------------------------------------------------------------------
  def test_subject_element(self):
    tpl = '''<html
   xmlns="http://www.w3.org/1999/xhtml"
   xmlns:email="http://xmlns.cadit.com/caditapp/email/1.0"
   >
   <head>
    <title><email:subject share="content">${message.title()}</email:subject></title>
   </head>
   <body>
    <p>${message}.</p>
   </body>
  </html>
'''
    chk = 'This Is A Test'
    eml = Manager(sender=StoredSender(), provider=StringProvider(tpl)).newEmail()
    eml['message'] = 'This is a test'
    self.assertEqual(eqstrip(chk), eqstrip(eml.getSubject()))
    # TODO: ensure share="content" works... (it doesn't currently...)

  #----------------------------------------------------------------------------
  def test_subject_entities(self):
    tpl = '''<html
   xmlns="http://www.w3.org/1999/xhtml"
   xmlns:email="http://xmlns.cadit.com/caditapp/email/1.0"
   >
   <head>
    <title email:subject="content">${message.title()}&nbsp;&mdash;&nbsp;And More!...</title>
   </head>
   <body>
    <h1>Foo &amp; Bar</h1>
    <p>${message}&nbsp;&mdash;&nbsp;and more!...</p>
   </body>
  </html>
'''
    schk = 'This & That -- And More!...'
    hchk = '''<html xmlns="http://www.w3.org/1999/xhtml">
    <head>
      <title>This &amp; That&#160;&#8212;&#160;And More!...</title>
    </head>
    <body>
    <h1>Foo &amp; Bar</h1>
    <p>This &amp; that&#160;&#8212;&#160;and more!...</p>
   </body>
  </html>
'''
    eml = Manager(sender=StoredSender(), provider=StringProvider(tpl)).newEmail()
    eml['message'] = 'This & that'
    self.assertEqual(schk, eml.getSubject())
    self.assertEqual(eqstrip(hchk), eqstrip(eml.getHtml()))

  #----------------------------------------------------------------------------
  def test_subject_attribute(self):
    tpl = '''<html
   xmlns="http://www.w3.org/1999/xhtml"
   xmlns:email="http://xmlns.cadit.com/caditapp/email/1.0"
   >
   <head>
    <title email:subject="content">${message.title()}</title>
   </head>
   <body>
    <p>${message}.</p>
   </body>
  </html>
'''
    chk = 'This Is A Test'
    eml = Manager(sender=StoredSender(), provider=StringProvider(tpl)).newEmail()
    eml['message'] = 'This is a test'
    self.assertEqual(eqstrip(chk), eqstrip(eml.getSubject()))

  #----------------------------------------------------------------------------
  def test_subject_maxlength(self):
    tpl = '''<html
 xmlns="http://www.w3.org/1999/xhtml"
 xmlns:email="http://xmlns.cadit.com/caditapp/email/1.0"
 >
 <head>
  <title><email:subject share="content">${message.title()}</email:subject></title>
 </head>
 <body>
  <p>${message}.</p>
 </body>
</html>
'''
    chk = 'This Is A Test [...]'
    man = Manager(sender=StoredSender(), provider=StringProvider(tpl))
    man.defaultMaxSubjectLength = 20
    eml = man.newEmail()
    eml['message'] = 'This is a test of capping the subject length'
    self.assertEqual(chk, eml.getSubject())
    chk2 = 'This [...]'
    eml2 = man.newEmail()
    eml2.maxSubjectLength = 10
    eml2['message'] = 'This is a test of capping the subject length'
    self.assertEqual(chk2, eml2.getSubject())

  #----------------------------------------------------------------------------
  def test_subject_snip(self):
    tpl = '''<html
 xmlns="http://www.w3.org/1999/xhtml"
 xmlns:email="http://xmlns.cadit.com/caditapp/email/1.0"
 >
 <head>
  <title><email:subject share="content">${message.title()}</email:subject></title>
 </head>
 <body>
  <p>${message}.</p>
 </body>
</html>
'''
    chk = 'This Is [...snip...]'
    man = Manager(sender=StoredSender(), provider=StringProvider(tpl))
    man.defaultMaxSubjectLength = 20
    man.defaultSnipIndicator = '[...snip...]'
    eml = man.newEmail()
    eml['message'] = 'This is a test of changing the snip indicator'
    self.assertEqual(chk, eml.getSubject())

  #----------------------------------------------------------------------------
  def test_deprecationwarning_workaround(self):
    html = '<html><body><p class="f">first</p><p>second</p></body></html>'
    css  = 'p{background:#666;}p.f{background:#042d5a;}'
    import warnings
    warnings.filterwarnings(
      "error",
      message='Call to deprecated method \'_[sg]etCSSValue\'. Use ``property.propertyValue`` instead.',
      category=DeprecationWarning,
      module='cssutils.css.cssstyledeclaration',
      lineno=598,
      )
    xdoc = emailprov.parseXml(html)
    xdoc = emailprov.inlineHtmlStyling(xdoc, css)
    out  = emailprov.serializeHtml(xdoc)
    self.assertEqual(eqstrip('''<html xmlns="http://www.w3.org/1999/xhtml">
  <body>
    <p class="f" style="background: #042d5a">first</p>
    <p style="background: #666">second</p>
  </body>
</html>
'''),
                     eqstrip(out))

  #----------------------------------------------------------------------------
  def test_simpleEmail_textOnly(self):
    tpl = '''<html
 xmlns="http://www.w3.org/1999/xhtml"
 xmlns:email="http://xmlns.cadit.com/caditapp/email/1.0"
 >
 <head>
  <email:header name="To">test@example.com</email:header>
  <email:header name="From" value="noreply@example.com">this-is-bogus</email:header>
  <email:env name="unused">feedback value</email:env>
  <title email:subject="content">${message.title()}</title>
 </head>
 <body>
  <p>${message}.</p>
  <p>Also sent to: <span email:header="CC">foo@example.com</span>.</p>
 </body>
</html>
'''
    tchk = '''This is a test.\n\nAlso sent to: foo@example.com.\n\n'''
    etchk = '''MIME-Version: 1.0
Content-Type: text/plain; charset="us-ascii"
Content-Transfer-Encoding: 7bit
From: noreply@example.com
CC: foo@example.com
To: test@example.com
Date: Fri, 13 Feb 2009 23:31:30 -0000
Message-ID: <1234567890@caditapp.email.test>
Subject: This Is A Test\n\n''' + tchk
    hchk = '''<html xmlns="http://www.w3.org/1999/xhtml">
  <head>
    <title>This Is A Test</title>
  </head>
  <body>
  <p>This is a test.</p>
  <p>Also sent to: <span>foo@example.com</span>.</p>
 </body>
</html>
'''
    eml = Manager(sender=StoredSender(), provider=StringProvider(tpl)).newEmail()
    #eml.setHeader('From', 'noreply@example.com')
    #eml.setHeader('To', 'test@example.com')
    # override the UNpredictable generated info...
    eml.setHeader('Date', 'Fri, 13 Feb 2009 23:31:30 -0000')
    eml.setHeader('Message-ID', '<1234567890@caditapp.email.test>')
    eml.include = ['text']
    eml['message'] = 'This is a test'
    eml.send()
    self.assertEqual(1, len(eml.manager.sender.emails))
    out = eml.manager.sender.emails[0]
    self.assertEqual(eqstrip(etchk), eqstrip(out['message']))
    self.assertEqual(eqstrip(tchk), eqstrip(eml.getText()))
    self.assertEqual(eqstrip(hchk), eqstrip(eml.getHtml()))
    self.assertEqual(sorted(['test@example.com', 'foo@example.com']),
                     sorted(out['recipients']))

  #----------------------------------------------------------------------------
  def test_cidCleanup(self):
    tpl = '''<html
 xmlns="http://www.w3.org/1999/xhtml"
 xmlns:email="http://xmlns.cadit.com/caditapp/email/1.0"
 >
 <head>
  <email:header name="To">test@example.com</email:header>
  <email:header name="From" value="noreply@example.com">this-is-bogus</email:header>
  <title email:subject="content">${message.title()}</title>
  <email:attachment name="smiley.png" encoding="base64" cid="true">
   dGhpcyBpcyBhIGJvZ3VzIGltYWdlCg==
  </email:attachment>
 </head>
 <body>
  <p>${message}.</p>
  <p>Also sent to: <span email:header="CC">foo@example.com</span>.</p>
  <p><img alt="smiley" src="cid:smiley.png"/></p>
 </body>
</html>
'''
    chk = '''This is a test.

Also sent to: foo@example.com.

[smiley]

'''
    eml = Manager(sender=StoredSender(), provider=StringProvider(tpl)).newEmail()
    eml['message'] = 'This is a test'
    self.assertEqual(eqstrip(chk), eqstrip(eml.getText()))

  #----------------------------------------------------------------------------
  def test_inlineStyling(self):
    tpl = '''<html
 xmlns="http://www.w3.org/1999/xhtml"
 xmlns:email="http://xmlns.cadit.com/caditapp/email/1.0"
 >
 <head>
  <email:header name="To">test@example.com</email:header>
  <email:header name="From" value="noreply@example.com">this-is-bogus</email:header>
  <title email:subject="content">${message.title()}</title>
<style type="text/css">
p > span { font-weight: bold; }
  </style>
 </head>
 <body>
  <p>${message}.</p>
  <p>Also sent to: <span email:header="CC">foo@example.com</span>.</p>
 </body>
</html>
'''
    tchk = '''This is a test.\n\nAlso sent to: foo@example.com.\n\n'''
    etchk = '''MIME-Version: 1.0
Content-Type: text/plain; charset="us-ascii"
Content-Transfer-Encoding: 7bit
From: noreply@example.com
CC: foo@example.com
To: test@example.com
Date: Fri, 13 Feb 2009 23:31:30 -0000
Message-ID: <1234567890@caditapp.email.test>
Subject: This Is A Test\n\n''' + tchk
    hchk = '''<html xmlns="http://www.w3.org/1999/xhtml">
  <head>
    <title>This Is A Test</title>
  </head>
  <body>
  <p>This is a test.</p>
  <p>Also sent to: <span style="font-weight: bold">foo@example.com</span>.</p>
 </body>
</html>
'''
    eml = Manager(sender=StoredSender(), provider=StringProvider(tpl)).newEmail()
    #eml.setHeader('From', 'noreply@example.com')
    #eml.setHeader('To', 'test@example.com')
    # override the UNpredictable generated info...
    eml.setHeader('Date', 'Fri, 13 Feb 2009 23:31:30 -0000')
    eml.setHeader('Message-ID', '<1234567890@caditapp.email.test>')
    eml.include = ['text']
    eml['message'] = 'This is a test'
    eml.send()
    #self.assertMultiLineEqual(etchk, eml.manager.sender.emails[0]['message'])
    #self.assertMultiLineEqual(tchk, eml.getText())
    self.assertEqual(eqstrip(hchk), eqstrip(eml.getHtml()))

  #----------------------------------------------------------------------------
  def test_email_rawSmtp(self):
    tpl = '''<html
 xmlns="http://www.w3.org/1999/xhtml"
 xmlns:email="http://xmlns.cadit.com/caditapp/email/1.0"
 >
 <head>
  <email:header name="To">test@example.com</email:header>
  <email:header name="From" value="noreply@example.com"/>
  <email:attachment name="smiley.png" encoding="base64" cid="true">
   iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAACXBIWXMAAAsTAAALEwEAmpwYAAAA
   IGNIUk0AAHolAACAgwAA+f8AAIDpAAB1MAAA6mAAADqYAAAXb5JfxUYAAAK0SURBVHjabJNfaJNX
   GMZ/X/IlaWub1NqWzLY6dYyiKAWxF0MdyC4KgjeObWy72G28EESam93sOlUZA82Nt2OMiWWwQmFg
   BXfVIQ6JqdY0sRi1TZfG/OmX7/vOOXm9KIRU+8KB9+F9zvPAe85jiQidtfrnxJCIJIAphEkBEFkU
   mEdIH774eKOTb3UKFO6emA6E+1LR0c/pHjxGV+wQAG61wNZGhurqAtqvJz/96snMBwK5O8dvxcbO
   Jvo//oJQpBdaLiJqm2SFINCFcuuU83+zWVhIH/3u6aW2wPLvx1KxsTPTw+MXQFdA/Y/oGrScbZtA
   D5YdhdAg2Ht5nZmlnL8/c+KH50kr++v4SDASKx757DKWWgPvFYv35kE0p05PAPDvP/+BZTN5bgoi
   I4gdJ7twDd+pjAa0NolYfALLySO1DNIosJJdZmUpD34T/CYrS3lWsstIo4DUMljNPAMjJ9HKJGyt
   zPk9PV20qllQFRDTXqr47o4XEq8BfhPxXfp6h1DKnLe1Nkcjpoy4b3ZcBhDl7o6VSyQEWptDtlKm
   qRtvwsHWVpv4zdXbAJjczV0xQEtvoJQJ2lqZQrNentgTbrWHeunGDuf3MYDjltDK5AJa6bnN8ltE
   +YjyWXz0st2/fzpnpVIFrcxcQCmTfrnewPguaI8DBw7y4GGRtfUqnuPgOQ5r61UePCwyFAuB9jC+
   y4vXVbQ2aUtE+Oun/tTIgD09vr8Lb+9ZNktFni0/p16vAdDXF+XIaIyP4kPYzQKZ1S1W192ZL1PN
   ZPsrz/7Ye2tsIJQ4vL+f7micQPQTLLt3e/u6QauWw6kUyb2q8qLkpb++7l36IEx/JLunw0FSB/fZ
   DEeD9PcEAKhsGUo1Q2FD4fqt5Lc/q5ld0wjw25VwXCCByJTAJNt5XhRhHkh//4te6+S/GwD2npI7
   ZK15EgAAAABJRU5ErkJggg==
  </email:attachment>
  <title email:subject="content">${message.title()}</title>
 </head>
 <body>
  <p>${message}.</p>
  <p><img alt="smiley" src="cid:smiley.png"/></p>
 </body>
</html>
'''
    schk = 'This And That'
    hchk = '''<html xmlns="http://www.w3.org/1999/xhtml">
  <head>
    <title>This And That</title>
  </head>
  <body>
  <p>This and that.</p>
  <p><img alt="smiley" src="cid:smiley.png"/></p>
 </body>
</html>
'''

    regex = re.compile('.*<email:attachment[^>]*>([^<]*)</email:attachment>.*', re.DOTALL)

    chk = '''Content-Type: multipart/alternative; boundary="==caditapp.email.test=="
MIME-Version: 1.0
Date: Fri, 13 Feb 2009 23:31:30 -0000
To: test@example.com
Message-ID: <1234567890@caditapp.email.test>
From: noreply@example.com
Subject: This And That

--==caditapp.email.test==
MIME-Version: 1.0
Content-Type: text/plain; charset="us-ascii"
Content-Transfer-Encoding: 7bit

This and that.

[smiley]


--==caditapp.email.test==
Content-Type: multipart/related; boundary="==caditapp.email.test-related=="
MIME-Version: 1.0

--==caditapp.email.test-related==
MIME-Version: 1.0
Content-Type: text/html; charset="us-ascii"
Content-Transfer-Encoding: 7bit

<html xmlns="http://www.w3.org/1999/xhtml">
  <head>
    <title>This And That</title>
  </head>
  <body>
  <p>This and that.</p>
  <p><img alt="smiley" src="cid:smiley.png"/></p>
 </body>
</html>

--==caditapp.email.test-related==
Content-Type: image/png; name="smiley.png"
MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Disposition: attachment
Content-ID: <smiley.png>

''' + regex.sub('\\1', tpl).replace('   ','').strip() + '''
--==caditapp.email.test-related==--
--==caditapp.email.test==--'''


    eml = Manager(sender=StoredSender(), provider=StringProvider(tpl)).newEmail()
    # override the UNpredictable generated info...
    eml.setHeader('Date', 'Fri, 13 Feb 2009 23:31:30 -0000')
    eml.setHeader('Message-ID', '<1234567890@caditapp.email.test>')
    eml.boundary = 'caditapp.email.test'
    eml['message'] = 'This and that'
    eml.send()
    self.assertEqual(schk, eml.getSubject())
    self.assertEqual(eqstrip(hchk), eqstrip(eml.getHtml()))
    self.assertEqual(eqstrip(chk), eqstrip(eml.manager.sender.emails[0]['message']))

  #----------------------------------------------------------------------------
  def test_reduce2ascii_is_needed(self):
    from .emailprov import reduce2ascii as r2a
    self.assertRaises(UnicodeEncodeError,
                      u'this \u21d2 that'.encode,
                      'ascii')
    r2a(u'this \u21d2 that').encode('ascii')

  #----------------------------------------------------------------------------
  def test_reduce2ascii(self):
    from .emailprov import reduce2ascii as r2a
    self.assertEqual('this => that', r2a(u'this \u21d2 that'))
    self.assertEqual('this <- that', r2a(u'this \u2190 that'))
    self.assertEqual('this <= that', r2a(u'this \u21e6 that'))
    self.assertEqual(
      '->=>->=>=>->|->->|->=>>->|=>~>->>->',
      r2a(u'\u2192\u21d2\u21ac\u21c9\u21e8\u21e5\u21e2\u21aa\u21a6\u21f6\u21a3\u21f0\u219d\u21a0\u21fe'))

  #----------------------------------------------------------------------------
  def test_bccHeader(self):
    tpl = '''<html
 lang="en"
 xml:lang="en"
 xmlns="http://www.w3.org/1999/xhtml"
 xmlns:email="http://xmlns.cadit.com/caditapp/email/1.0"
 >
 <head>
  <title email:subject="content">${message.title()}</title>
  <email:header name="to">rcpt@example.com</email:header>
  <email:header name="from">mailfrom@example.com</email:header>
 </head>
 <body><p>${message}.</p></body>
</html>
'''
    sender  = StoredSender()
    manager = Manager(sender=sender, provider=StringProvider(tpl))
    eml = manager.newEmail()
    eml['message'] = 'test'
    eml.setHeader('BcC', 'bcc@example.com')
    # override the UNpredictable generated info...
    eml.setHeader('date', 'Fri, 13 Feb 2009 23:31:30 -0000')
    eml.setHeader('message-id', '<1234567890@caditapp.email.test>')
    eml.boundary = 'caditapp.email.test'
    eml.send()
    self.assertEqual(1, len(sender.emails))
    out = sender.emails[0]
    self.assertEqual(sorted(['mailfrom', 'recipients', 'message']), sorted(out.keys()))
    self.assertEqual('mailfrom@example.com', out['mailfrom'])
    self.assertEqual(sorted(['rcpt@example.com', 'bcc@example.com']), sorted(out['recipients']))
    # ensure that the 'bcc' headers is stripped out of the output
    chk = '''Content-Type: multipart/alternative; boundary="==caditapp.email.test=="
MIME-Version: 1.0
From: mailfrom@example.com
To: rcpt@example.com
Date: Fri, 13 Feb 2009 23:31:30 -0000
Message-ID: <1234567890@caditapp.email.test>
Subject: Test

--==caditapp.email.test==
MIME-Version: 1.0
Content-Type: text/plain; charset="us-ascii"
Content-Transfer-Encoding: 7bit

test.


--==caditapp.email.test==
Content-Type: multipart/related; boundary="==caditapp.email.test-related=="
MIME-Version: 1.0

--==caditapp.email.test-related==
MIME-Version: 1.0
Content-Type: text/html; charset="us-ascii"
Content-Transfer-Encoding: 7bit

<html xmlns="http://www.w3.org/1999/xhtml" lang="en" xml:lang="en">
  <head>
    <title>Test</title>
  </head>
  <body>
    <p>test.</p>
  </body>
</html>

--==caditapp.email.test-related==--
--==caditapp.email.test==--'''
    self.assertEqual(eqstrip(chk), eqstrip(out['message']))

  #----------------------------------------------------------------------------
  def test_managerBccHeader(self):
    tpl = '''<html
 xmlns="http://www.w3.org/1999/xhtml"
 xmlns:email="http://xmlns.cadit.com/caditapp/email/1.0"
 >
 <head>
  <title email:subject="content">${message.title()}</title>
  <email:header name="to">rcpt@example.com</email:header>
  <email:header name="from">mailfrom@example.com</email:header>
 </head>
 <body><p>${message}.</p></body>
</html>
'''
    sender  = StoredSender()
    class MyManager(Manager):
      def updateHeaders(self, emailobj, headers):
        super(MyManager, self).updateHeaders(emailobj, headers)
        headers['bcc'] = 'bcc@example.com'
    manager = MyManager(sender=sender, provider=StringProvider(tpl))
    eml = manager.newEmail()
    eml['message'] = 'test'
    # override the UNpredictable generated info...
    eml.setHeader('date', 'Fri, 13 Feb 2009 23:31:30 -0000')
    eml.setHeader('message-id', '<1234567890@caditapp.email.test>')
    eml.send()
    self.assertEqual(1, len(sender.emails))
    out = sender.emails[0]
    self.assertEqual(sorted(['rcpt@example.com', 'bcc@example.com']), sorted(out['recipients']))

  # #----------------------------------------------------------------------------
  # def test_multiBccHeader(self):
  #   # todo
  #   pass

  #----------------------------------------------------------------------------
  def test_env_extraction(self):
    tpl = '''<html
 xmlns="http://www.w3.org/1999/xhtml"
 xmlns:email="http://xmlns.cadit.com/caditapp/email/1.0"
 >
 <head>
  <email:header name="To">test@example.com</email:header>
  <email:header name="From" value="noreply@example.com">this-is-bogus</email:header>
  <email:env name="env-name">feedback env-value</email:env>
  <title email:subject="content">${message.title()}</title>
 </head>
 <body>
  <p>${message}.</p>
  <p>Also sent to: <span email:header="CC">foo@example.com</span>.</p>
 </body>
</html>
'''

    eml = Manager(sender=StoredSender(), provider=StringProvider(tpl)).newEmail()
    self.assertEqual(eml.getEnv('env-name'), 'feedback env-value')
    self.assertEqual(eml.getEnvs(), dict([['env-name', 'feedback env-value']]))

#------------------------------------------------------------------------------
if __name__ == '__main__':
  unittest.main() # pragma: no cover

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2012/11/20
# copy: (C) Copyright 2012 Cadit Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import os.path, yaml, email, pkg_resources, mako
from .emailprov import Email
from .sender import SmtpSender

import logging
log = logging.getLogger(__name__)

#------------------------------------------------------------------------------
__all__     = ['Manager', 'Provider', 'StringProvider', 'FileProvider', 'PkgProvider']

#------------------------------------------------------------------------------
class NoSuchResource(Exception): pass

#------------------------------------------------------------------------------
class Provider(object):
  '''
  Abstract class that defines the interface for a template/object
  provider to the :class:`caditapp.email.Manager` object. This class
  provides two types of resources: templates (via :meth:`getTemplate`,
  which must conform to the mako template API) and maps (via
  :meth:`getMap`, which must conform to the dict API).

  The following named templates can be defined (only the first of
  which is required)::

    None            # base template
    "text"          # text-only template (default: base template)
    "html"          # html-only template (default: base template)
    "subject"       # subject template (default: dynamic value from base template)

  Unless :meth:`getTemplate` is overriden, these are simply requested
  from :meth:`get` with ``.mako`` suffixed to the name and parsed as
  mako templates.

  The following named maps can be defined (none of which are required)::

    "headers"       # static map of headers (default: dynamic values from base template)
    "attachments"   # static table of attachments (default: dynamic values from base template)

  Unless :meth:`getMap` is overriden, these are simply requested from
  :meth:`get` with ``.yaml`` suffixed to the name and parsed as YAML
  documents.
  '''
  def __init__(self, lookup=None, **kw):
    super(Provider, self).__init__(**kw)
    self.lookup = lookup
  def get(self, name=None, default=None):
    '''
    Return the content for the specified `name` named resource. If not
    found or not available, `default` should be returned.
    '''
    raise NotImplementedError()
  def __getitem__(self, name):
    ret = self.get(name)
    if ret is not None:
      return ret
    raise NoSuchResource(name)
  def getTemplate(self, name=None, default=None):
    '''
    Return the mako template named `name` from the current provider.
    '''
    dat = self.get(name is not None and name + '.mako' or None, None)
    if dat is None:
      return default
    return self._makeTemplate(dat)
  def _makeTemplate(self, data):
    '''
    Generates a mako template from the specified `data`.
    '''
    return mako.template.Template(data, lookup=self.lookup, default_filters=['h'])
  def getMap(self, name, default=None):
    '''
    Return the mapping object (i.e. dict) named `name` from the
    current provider.
    '''
    dat = self.get(name + '.yaml', None)
    if dat is None:
      return default
    return yaml.load(dat)
  def getSubProvider(self, name):
    '''
    Return a provider for resource named `name` that is subsidiary to
    the current provider. The default implementation simply returns a
    :class:`caditapp.email.PrefixProvider` instance relative to the
    current provider.
    '''
    if name is None:
      return self
    return PrefixProvider(name, self)

#------------------------------------------------------------------------------
class PrefixProvider(Provider):
  '''
  A :class:`caditapp.email.Provider` implementation that simply
  prefixes any requested template or map with `name` (provided in the
  constructor) and a period (".") from the `provider` provided in the
  constructor.
  '''
  def __init__(self, name, provider, **kw):
    super(PrefixProvider, self).__init__(**kw)
    self.name     = name
    self.provider = provider
  def get(self, name=None, default=None):
    if name is None:
      return self.provider.get(self.name + '.mako', default)
    return self.provider.get(self.name + '.' + name, default)
  def getTemplate(self, name=None, default=None):
    if name is None:
      return self.provider.getTemplate(self.name, default)
    return self.provider.getTemplate(self.name + '.' + name, default)
  def getMap(self, name, default=None):
    return self.provider.getMap(self.name + '.' + name, default)

#------------------------------------------------------------------------------
class StringProvider(Provider):
  '''
  A simple :class:`caditapp.email.Provider` implementation that simply
  returns the *content* as the base template. All other resource
  requests will be responded to with the default value, or None if not
  specified.
  '''
  def __init__(self, content, **kw):
    super(StringProvider, self).__init__(**kw)
    self.content = content
  def get(self, name=None, default=None):
    if name is None: return self.content
    return default

#------------------------------------------------------------------------------
class FileProvider(Provider):
  '''
  Loads templates and maps from the filesystem. With `basedir` set to
  ``\'/path/to/base\'`` and loading email named ``welcome``, the
  following files will be loaded (where all except ``welcome.mako``
  are optional)::

    -- /path/to/base/
       |-- welcome.mako                 # base template
       |-- welcome.text.mako            # optional template for text content
       |-- welcome.html.mako            # optional template for html content
       |-- welcome.subject.mako         # optional template for subject
       |-- welcome.headers.yaml         # optional map for headers
       `-- welcome.attachments.yaml     # optional map for attachments

  See :class:`caditapp.email.Provider` for more details on templates
  and maps that can be requested and how they get resolved by default.

  The `defname` provided in the constructor is only used if an email
  is requested without a name, in which case `defname` is substituted
  for the name.

  NOTE: an issue with the current implementation is that XInclude\'s
  may not be resolved correctly. See the
  :class:`caditapp.email.MakoFileProvider` for a subclass that may
  better address your needs.
  '''
  def __init__(self, basedir, defname=None, **kw):
    super(FileProvider, self).__init__(**kw)
    self.basedir = basedir
    self.defname = defname
  def get(self, name=None, default=None):
    if name is None:
      # in theory, this should never happen...
      return self.get(os.path.join(self.basedir, self.defname + '.mako'), default)
    return self.readFile(os.path.join(self.basedir, name), default)
  def readFile(self, filename, default=None):
    try: fd = open(filename, 'rb')
    except: return default
    try: ret = fd.read()
    except: ret = default
    else: fd.close()
    return ret
  def getSubProvider(self, name):
    return super(FileProvider, self).getSubProvider(name or self.defname)

#------------------------------------------------------------------------------
class PkgProvider(FileProvider):
  '''
  Identical to :class:`caditapp.email.FileProvider` except it uses
  ``pkg_resources`` to load resources.
  '''
  def __init__(self, spec, defname=None, **kw):
    self.pkgname, root = spec.split(':', 1)
    return super(PkgProvider, self).__init__(root, defname, **kw)
  def readFile(self, filename, default=None):
    if not pkg_resources.resource_exists(self.pkgname, filename):
      return default
    try:
      ret = pkg_resources.resource_stream(self.pkgname, filename).read()
    except Exception:
      log.exception('could not load resource "%s:%s"', self.pkgname, filename)
      return default
    return ret
  def getSubProvider(self, name):
    return super(PkgProvider, self).getSubProvider(name or self.defname)

#------------------------------------------------------------------------------
class Manager(object):
  '''
  A ``caditapp.email.Manager`` object is the main clearinghouse for
  generating templatized emails. The main objective is that it
  provides access to :class:`caditapp.email.Email` objects. It
  requires access to a :class:`caditapp.email.Provider` object that
  provides templates, and objects referenced by the templates (such as
  embedded images).

  Constructor parameters:

  :param provider:
    the default :class:`caditapp.email.Provider` object (which
    provides the raw templates for the emails) for this Manager. if no
    Provider is provided, then the ``newEmail`` method will not be
    useful.

  :param sender:
    the :class:`caditapp.email.Sender` implementation that will
    actually initiate the transfer of an email to a delivery
    destination. if not provided, a default
    :class:`caditapp.email.SmtpSender` will be instantiated.
  '''

  # TODO: since the manager has to be created in order to create an
  #       Email object, i can cache templates... i should take
  #       advantage of that!...

  #----------------------------------------------------------------------------
  def __init__(self, provider=None, sender=None):
    self.provider = provider
    self.sender   = sender or SmtpSender()
    self.context  = 'caditapp.email'
    self.defaultBoundary           = None
    self.defaultTextRenderEncoding = None
    self.defaultHtmlRenderEncoding = None
    self.defaultTextEncoding       = None
    self.defaultHtmlEncoding       = None
    self.defaultTransferEncoding   = None
    self.defaultEncoding           = None
    self.defaultMaxSubjectLength   = None
    self.defaultSnipIndicator      = '[...]'

  #----------------------------------------------------------------------------
  def newEmail(self, name=None):
    '''
    Creates and returns a new :class:`caditapp.email.Email` object
    that will be based on the templates named `name` from the current
    :class:`caditapp.email.Provider` object. Note that for most
    providers, a ``None`` `name` is valid and refers to the default
    template.
    '''
    return Email(manager=self, provider=self.provider.getSubProvider(name))

  #----------------------------------------------------------------------------
  def newEmailFromTemplate(self, template):
    '''
    Creates and returns a new :class:`caditapp.email.Email` object
    that will be based on the provided string `template`. A
    :class:`caditapp.email.StringProvider` will be created to provide
    the template to the :class:`caditapp.email.Email` object.
    Although the returned object will not use the manager\'s Provider,
    it *will* use the manager\'s sender.
    '''
    return Email(manager=self, provider=StringProvider(template))

  #----------------------------------------------------------------------------
  def updateHeaders(self, emailobj, headers):
    defaultHeaders = {
      'Date':        lambda: email.Utils.formatdate(),
      'Message-ID':  lambda: email.Utils.make_msgid(self.context),
      }
    for name, call in defaultHeaders.items():
      if headers.has_key(name):
        continue
      headers[name] = call()

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# lib:  templatealchemy.mako
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2013/07/03
# copy: (C) Copyright 2013 Cadit Health Inc., All Rights Reserved.
#------------------------------------------------------------------------------

from __future__ import absolute_import

import mako.template
from mako.lookup import TemplateLookup
from templatealchemy import api, util, engine

#------------------------------------------------------------------------------
def loadRenderer(spec=None):
  return MakoRenderer(spec)

#------------------------------------------------------------------------------
class TaLookup(TemplateLookup):
  # TODO: this is not very efficient in the case that building a source
  #       is expensive...
  def __init__(self, source, renderer):
    self.source   = source
    self.renderer = renderer
  def adjust_uri(self, uri, relativeto):
    return self.source.resolveUri(uri, relativeto)
  def get_template(self, uri):
    src = engine.loadSource(uri)
    return self.ta_template(text=src.get('').read(), uri=uri)
  def ta_template(self, **kw):
    kw['lookup'] = self
    kw['default_filters'] = self.renderer.filters
    return mako.template.Template(**kw)

#------------------------------------------------------------------------------
class MakoRenderer(api.Renderer):

  #----------------------------------------------------------------------------
  def __init__(self, spec, *args, **kw):
    super(MakoRenderer, self).__init__(self.ns('mako', spec), *args, **kw)
    # TODO: expose control of `mako.template.Template()` args/kwargs...
    self.spec = spec
    self.filters = ['h']

  #----------------------------------------------------------------------------
  def render(self, context, stream, params):
    if context.makoLookup is None:
      context.makoLookup = TaLookup(context.template.source, self)
    return context.makoLookup.ta_template(
      text = stream.read(),
      uri  = context.template.source.uri,
      ).render(**params)

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# lib:  templatealchemy.cli
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2013/07/06
# copy: (C) Copyright 2013 Cadit Health Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import sys, argparse, yaml, os.path
from . import engine, stream

#------------------------------------------------------------------------------
def main(args=None, output=None):

  cli = argparse.ArgumentParser(
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description='''\
Command-line template evaluator using the TemplateAlchemy template abstraction
layer.
''',
    epilog='''\
Examples:

  # using shell pipelines with mustache rendering
  $ echo 'Hello, {{name}}!' | %(prog)s -y "{name: World}" -r mustache
  Hello, World!

  # using a file source with mustache rendering
  $ echo 'Hello, {{name}}!' > sample.text
  $ %(prog)s -y "{name: World}" -r mustache ./sample.text
  Hello, World!

  # using a sqlite3 source and mako rendering (the default)
  $ sqlite3 templates.db 'CREATE TABLE template (name VARCHAR, format VARCHAR, content TEXT);'
  $ sqlite3 templates.db 'INSERT INTO template VALUES ("foo/bar", "text", "Hello, ${name}!");'
  $ %(prog)s -y "{name: World}" -n foo/bar -f text sqlalchemy:sqlite:///templates.db
  Hello, World!
''')

  cli.add_argument(
    '-v', '--verbose',
    action='count',
    help='enable verbose output (multiple invocations increase verbosity)')

  cli.add_argument(
    '-n', '--name', metavar='NAME',
    default=None, action='store',
    help='set the template name; if omitted, the root template will be used')

  cli.add_argument(
    '-f', '--format', metavar='FORMAT',
    default=None, action='store',
    help='set the output format; if omitted, it will use the template\'s'
    ' default format')

  cli.add_argument(
    '-p', '--param', metavar='NAME=VALUE',
    default=[], action='append',
    help='set a template variable where `VALUE` is taken as a literal'
    ' string (overrides any values set in `--params`)')

  cli.add_argument(
    '-y', '--params', metavar='YAML',
    default=[], action='append',
    help='specifies a YAML-encoded dictionary as template variables;'
    ' if the value starts with the letter "@", the rest is the file name to'
    ' read as a YAML structure. If the value is exactly a dash ("-"), then'
    ' the YAML structure is read from STDIN. In all cases, the YAML'
    ' structure must be a dictionary')

  cli.add_argument(
    '-r', '--renderer', metavar='SPEC',
    default='mako',
    help='sets the rendering driver (default: %(default)r)')

  cli.add_argument(
    'source', metavar='SOURCE',
    nargs='?',
    help='the template source; if exactly a dash ("-") or omitted, then'
    ' the template is read from STDIN. If both params and template are read'
    ' from STDIN, then params is read first, followed by an EOF, then the'
    ' template. If the source lacks a colon (":") and it points to a normal'
    ' file, the template is read from the file. Otherwise, `source` is'
    ' interpreted as a source specification (i.e. "TYPE[:OPTIONS]")')

  options = cli.parse_args(args)

  params = dict()

  for yparam in options.params:
    if yparam == '-':
      yparam = sys.stdin.read()
    elif yparam.startswith('@'):
      with open(yparam[1:], 'rb') as fp:
        yparam = fp.read()
    try:
      yparam = yaml.load(yparam)
    except Exception:
      cli.error('could not parse YAML expression: %r'
                % (yparam,))
    if not isinstance(yparam, dict):
      cli.error('"--params" expressions must resolve to dictionaries')
    params.update(yparam)

  for kparam in options.param:
    if '=' not in kparam:
      cli.error('"--param" expressions must be in the KEY=VALUE format')
    key, value = kparam.split('=', 1)
    params[key] = value

  options.params = params
  output = output or sys.stdout

  if options.source is None or options.source == '-':
    options.source = stream.StreamSource(sys.stdin)
  elif ':' not in options.source and os.path.isfile(options.source):
    options.source = stream.StreamSource(open(options.source, 'rb'))

  template = engine.Template(
    source=options.source,
    renderer=options.renderer,
    )

  if options.name is not None:
    template = template.getTemplate(options.name)

  output.write(template.render(options.format, options.params))

  return 0

#------------------------------------------------------------------------------
if __name__ == '__main__':
  sys.exit(main())

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# lib:  templatealchemy.util
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2013/07/03
# copy: (C) Copyright 2013 Cadit Health Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import inspect

#------------------------------------------------------------------------------
def resolve(spec):
  spec = spec.split('.')
  used = spec.pop(0)
  found = __import__(used)
  for cur in spec:
    used += '.' + cur
    try:
      found = getattr(found, cur)
    except AttributeError:
      __import__(used)
      found = getattr(found, cur)
  return found

#------------------------------------------------------------------------------
class adict(dict):
  def __getattr__(self, key):
    return self.get(key, None)
  def __setattr__(self, key, value):
    self[key] = value
    return self
  def __delattr__(self, key):
    if key in self:
      del self[key]
    return self
  def update(self, *args, **kw):
    args = [e for e in args if e]
    dict.update(self, *args, **kw)
    return self
  @staticmethod
  def __dict2adict__(subject, recursive=False):
    if isinstance(subject, list):
      if not recursive:
        return subject
      return [adict.__dict2adict__(val, True) for val in subject]
    if not isinstance(subject, dict):
      return subject
    ret = adict(subject)
    if not recursive:
      return ret
    for key, val in ret.items():
      ret[key] = adict.__dict2adict__(val, True)
    return ret

#------------------------------------------------------------------------------
def callingPkgName(ignore=None):
  if ignore is None:
    ignore = []
  elif isinstance(ignore, basestring):
    ignore = [ignore]
  else:
    ignore = ignore[:]
  ignore.append('templatealchemy')
  ignore.append('ta')

  stack = inspect.stack()
  record = None
  try:
    for record in stack:
      if not record or not record[0]:
        continue
      mod = inspect.getmodule(record[0])
      if not mod:
        continue
      mod = getattr(mod, '__package__', None)
      if mod not in ignore:
        return mod
    return None
  finally:
    del record
    del stack

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# lib:  templatealchemy.api
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2013/07/03
# copy: (C) Copyright 2013 Cadit Health Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import os.path

#------------------------------------------------------------------------------
class Source(object):

  #----------------------------------------------------------------------------
  def __init__(self, uri, *args, **kw):
    self.uri = uri

  #----------------------------------------------------------------------------
  def get(self, format):
    '''
    Returns the source content stream for the current template source
    for the specified `format`. The returned object must be a
    file-like object supporting read access.
    '''
    raise NotImplementedError()

  #----------------------------------------------------------------------------
  def getSource(self, name):
    '''
    Returns a subsidiary source template, relative to the current
    template, with the specified `name`. This is seen as a hierchical
    relationship, and is typically represented as a slash ('/')
    delimited path.
    '''
    raise NotImplementedError()

  #----------------------------------------------------------------------------
  def getFormats(self):
    '''
    Returns a list of all the available formats for this source.
    '''
    raise NotImplementedError()

  #----------------------------------------------------------------------------
  def getRelated(self, name):
    '''
    Returns a content stream for the related object `name` that is
    relative to the current template. Typically this is used for
    meta-information *spec* definitions using the "!include" or
    "!include-raw" directives. As with :meth:`get`, the returned
    object must be a file-like object supporting read access.
    '''
    raise NotImplementedError()

  #----------------------------------------------------------------------------
  def ns(self, scheme, spec):
    if not spec:
      return scheme
    return scheme + ':' + spec

  #----------------------------------------------------------------------------
  def resolveUri(self, uri, base=None):
    return os.path.normpath(
      os.path.join(os.path.dirname(base or self.uri), uri))

#------------------------------------------------------------------------------
class Renderer(object):

  #----------------------------------------------------------------------------
  def __init__(self, uri, *args, **kw):
    self.uri = uri

  #----------------------------------------------------------------------------
  def render(self, context, stream, params):
    '''
    Renders the given template data `stream` (as a read-access
    file-like object) to serialized rendered output. The given
    `params` provide variables that are typically passed to the
    template using template-specific mechanisms.

    todo: update this when the time comes:

    `context` is a reserved parameter that is intended to enable
    cross-driver optimizations, but has not been defined at this
    point.
    '''
    raise NotImplementedError()

  #----------------------------------------------------------------------------
  def ns(self, scheme, spec):
    if not spec:
      return scheme
    return scheme + ':' + spec

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------
